import re
import warnings

import pyarrow as pa  # type: ignore[import-untyped]
import pytest

from yads.converters import PyArrowConverter, PyArrowConverterConfig
from yads.constraints import NotNullConstraint
from yads.spec import YadsSpec, Column, Field
from yads.types import (
    YadsType,
    String,
    Integer,
    Float,
    Decimal,
    Boolean,
    Binary,
    Date,
    TimeUnit,
    Time,
    Timestamp,
    TimestampTZ,
    TimestampLTZ,
    TimestampNTZ,
    Duration,
    IntervalTimeUnit,
    Interval,
    Array,
    Struct,
    Map,
    JSON,
    Geometry,
    Geography,
    UUID,
    Void,
    Variant,
    Tensor,
)
from yads.exceptions import (
    UnsupportedFeatureError,
    ValidationWarning,
    ConverterConfigError,
)


# fmt: off
# %% Types
class TestPyArrowConverterTypes:
    @pytest.mark.parametrize(
        "yads_type, expected_pa_type, expected_warning",
        [
            (String(), pa.string(), None),
            (String(length=255), pa.string(), "length constraint will be lost"),
            (Integer(bits=8), pa.int8(), None),
            (Integer(bits=16), pa.int16(), None),
            (Integer(bits=32), pa.int32(), None),
            (Integer(bits=64), pa.int64(), None),
            (Integer(bits=8, signed=False), pa.uint8(), None),
            (Integer(bits=16, signed=False), pa.uint16(), None),
            (Integer(bits=32, signed=False), pa.uint32(), None),
            (Integer(bits=64, signed=False), pa.uint64(), None),
            (Float(bits=16), pa.float16(), None),
            (Float(bits=32), pa.float32(), None),
            (Float(bits=64), pa.float64(), None),
            (Decimal(), pa.decimal128(38, 18), None),
            (Decimal(precision=10, scale=2), pa.decimal128(10, 2), None),
            (Decimal(precision=10, scale=2, bits=128), pa.decimal128(10, 2), None),
            (Boolean(), pa.bool_(), None),
            (Binary(), pa.binary(), None),
            (Binary(length=8), pa.binary(8), None),
            (Date(), pa.date32(), None),
            (Date(bits=32), pa.date32(), None),
            (Date(bits=64), pa.date64(), None),
            (Time(), pa.time32("ms"), None),  # default unit ms -> time32
            (Time(unit=TimeUnit.S), pa.time32("s"), None),
            (Time(unit=TimeUnit.MS), pa.time32("ms"), None),
            (Time(unit=TimeUnit.US), pa.time64("us"), None),
            (Time(unit=TimeUnit.NS), pa.time64("ns"), None),
            (Time(bits=32, unit=TimeUnit.S), pa.time32("s"), None),
            (Time(bits=64, unit=TimeUnit.US), pa.time64("us"), None),
            (Timestamp(), pa.timestamp("ns"), None),
            (Timestamp(unit=TimeUnit.S), pa.timestamp("s"), None),
            (Timestamp(unit=TimeUnit.MS), pa.timestamp("ms"), None),
            (Timestamp(unit=TimeUnit.US), pa.timestamp("us"), None),
            (Timestamp(unit=TimeUnit.NS), pa.timestamp("ns"), None),
            (TimestampTZ(), pa.timestamp("ns", tz="UTC"), None),
            (TimestampTZ(unit=TimeUnit.S), pa.timestamp("s", tz="UTC"), None),
            (TimestampLTZ(), pa.timestamp("ns", tz=None), None),
            (TimestampNTZ(), pa.timestamp("ns", tz=None), None),
            (Duration(), pa.duration("ns"), None),
            (Interval(interval_start=IntervalTimeUnit.DAY), pa.month_day_nano_interval(), None),
            (Array(element=Integer()), pa.list_(pa.int32()), None),
            (Array(element=String(), size=2), pa.list_(pa.string(), list_size=2), None),
            (
                Struct(
                    fields=[
                        Field(name="a", type=Integer()),
                        Field(name="b", type=String()),
                    ]
                ),
                pa.struct([
                    pa.field("a", pa.int32()),
                    pa.field("b", pa.string()),
                ]),
                None,
            ),
            (Map(key=String(), value=Integer()), pa.map_(pa.string(), pa.int32()), None),
            (JSON(), pa.json_(storage_type=pa.utf8()) if hasattr(pa, 'json_') else pa.string(), "JSON type" if not hasattr(pa, 'json_') else None),
            (Geometry(), pa.string(), "PyArrowConverter does not support type: geometry for 'col1'."),
            (Geometry(srid=4326), pa.string(), "PyArrowConverter does not support type: geometry(srid=4326) for 'col1'."),
            (Geography(), pa.string(), "PyArrowConverter does not support type: geography for 'col1'."),
            (Geography(srid=4326), pa.string(), "PyArrowConverter does not support type: geography(srid=4326) for 'col1'."),
            (UUID(), pa.uuid() if hasattr(pa, 'uuid') else pa.string(), "UUID type" if not hasattr(pa, 'uuid') else None),
            (Void(), pa.null(), None),
            (Variant(), pa.string(), "PyArrowConverter does not support type: variant for 'col1'."),
            (Tensor(element=Integer(bits=32), shape=(10, 20)), pa.fixed_shape_tensor(pa.int32(), [10, 20]), None),
            (Tensor(element=Float(bits=64), shape=(5, 10, 15)), pa.fixed_shape_tensor(pa.float64(), [5, 10, 15]), None),
            (Tensor(element=String(), shape=(100,)), pa.fixed_shape_tensor(pa.string(), [100]), None),
        ],
    )
    def test_convert_type(
        self,
        yads_type: YadsType,
        expected_pa_type: pa.DataType,
        expected_warning: str | None,
    ):
        spec = YadsSpec(
            name="test_spec",
            version="1.0.0",
            columns=[Column(name="col1", type=yads_type)],
        )
        # Set fallback_type explicitly for unsupported types
        config = PyArrowConverterConfig(fallback_type=pa.string())
        converter = PyArrowConverter(config)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            schema = converter.convert(spec, mode="coerce")

        # Assert converted schema
        assert schema.names == ["col1"]
        assert schema.field("col1").type == expected_pa_type
        assert schema.field("col1").nullable is True

        # Assert warnings for unsupported types
        if expected_warning is not None:
            assert len(w) == 1
            assert issubclass(w[0].category, ValidationWarning)
            assert expected_warning in str(w[0].message)
        else:
            assert len(w) == 0

    def test_non_nullable_columns_and_nested_fields(self):
        nested = Struct(
            fields=[
                Field(name="x", type=Integer(), constraints=[NotNullConstraint()]),
                Field(name="y", type=String()),
            ]
        )
        spec = YadsSpec(
            name="test_spec",
            version="1.0.0",
            columns=[
                Column(
                    name="id", type=Integer(), constraints=[NotNullConstraint()]
                ),
                Column(name="struct", type=nested),
                Column(
                    name="arr",
                    type=Array(element=Struct(fields=[Field("z", Integer())])),
                ),
            ],
        )

        schema = PyArrowConverter().convert(spec)

        id_field = schema.field("id")
        assert id_field.nullable is False
        assert id_field.type == pa.int32()

        struct_field = schema.field("struct")
        assert struct_field.nullable is True
        assert pa.types.is_struct(struct_field.type)

        # Check nested struct fields' nullability
        struct_struct = struct_field.type
        assert struct_struct.num_fields == 2
        assert struct_struct.field("x").nullable is False
        assert struct_struct.field("x").type == pa.int32()
        assert struct_struct.field("y").nullable is True
        assert struct_struct.field("y").type == pa.string()

        # Array of struct with nested field
        arr_field = schema.field("arr")
        assert pa.types.is_list(arr_field.type)
        elem_struct = arr_field.type.value_type
        assert pa.types.is_struct(elem_struct)
        assert elem_struct.field("z").type == pa.int32()

    def test_large_string_binary_and_list_flags(self):
        spec = YadsSpec(
            name="t",
            version="1.0.0",
            columns=[
                Column(name="s", type=String()),
                Column(name="b", type=Binary()),
                Column(name="l", type=Array(element=String())),
            ],
        )
        config = PyArrowConverterConfig(
            use_large_string=True,
            use_large_binary=True,
            use_large_list=True,
        )
        schema = PyArrowConverter(config).convert(spec)

        assert schema.field("s").type == pa.large_string()
        assert schema.field("b").type == pa.large_binary()
        assert schema.field("l").type == pa.large_list(pa.large_string())

    def test_decimal_precision_coercion_and_raise(self):
        spec = YadsSpec(
            name="t",
            version="1.0.0",
            columns=[Column(name="d", type=Decimal(precision=39, scale=2, bits=128))],
        )

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            schema = PyArrowConverter().convert(spec, mode="coerce")

        assert schema.field("d").type == pa.decimal256(39, 2)
        assert len(w) == 1
        assert issubclass(w[0].category, ValidationWarning)
        assert "precision > 38 is incompatible" in str(w[0].message)

        with pytest.raises(UnsupportedFeatureError):
            PyArrowConverter(PyArrowConverterConfig(mode="raise")).convert(spec)

    def test_time_bits_unit_mismatch_coercion(self):
        # time32 with us -> coerced to time64
        spec1 = YadsSpec(
            name="t1",
            version="1.0.0",
            columns=[Column(name="t", type=Time(bits=32, unit=TimeUnit.US))],
        )
        with warnings.catch_warnings(record=True) as w1:
            warnings.simplefilter("always")
            schema1 = PyArrowConverter().convert(spec1, mode="coerce")
        assert schema1.field("t").type == pa.time64("us")
        assert len(w1) == 1
        assert issubclass(w1[0].category, ValidationWarning)
        assert "time32 supports only 's' or 'ms'" in str(w1[0].message)

        # time64 with ms -> coerced to time32
        spec2 = YadsSpec(
            name="t2",
            version="1.0.0",
            columns=[Column(name="t", type=Time(bits=64, unit=TimeUnit.MS))],
        )
        with warnings.catch_warnings(record=True) as w2:
            warnings.simplefilter("always")
            schema2 = PyArrowConverter().convert(spec2, mode="coerce")
        assert schema2.field("t").type == pa.time32("ms")
        assert len(w2) == 1
        assert issubclass(w2[0].category, ValidationWarning)
        assert "time64 supports only 'us' or 'ns'" in str(w2[0].message)

    def test_schema_and_field_metadata_coercion(self):
        field_meta = {"a": 1, "b": {"x": True}}
        spec = YadsSpec(
            name="t",
            version="1.0.0",
            metadata={"owner": "data-eng", "cfg": {"retries": 3}},
            columns=[
                Column(name="c", type=String(), metadata=field_meta),
            ],
        )
        schema = PyArrowConverter().convert(spec)

        # Schema-level metadata comes back as bytes -> bytes
        raw_schema_meta = schema.metadata or {}
        decoded_schema_meta = {k.decode(): v.decode() for k, v in raw_schema_meta.items()}
        assert decoded_schema_meta.get("owner") == "data-eng"
        assert decoded_schema_meta.get("cfg") == "{\"retries\":3}"

        # Field metadata is attached to the field
        field = schema.field("c")
        raw_field_meta = field.metadata or {}
        decoded_field_meta = {k.decode(): v.decode() for k, v in raw_field_meta.items()}
        assert decoded_field_meta.get("a") == "1"
        assert decoded_field_meta.get("b") == "{\"x\":true}"

    @pytest.mark.parametrize(
        "yads_type, type_name",
        [
            (Geometry(), "geometry"),
            (Geometry(srid=4326), "geometry(srid=4326)"),
            (Geography(), "geography"),
            (Geography(srid=4326), "geography(srid=4326)"),
            (Variant(), "variant"),
        ],
    )
    def test_raise_mode_for_unsupported_types(self, yads_type: YadsType, type_name: str):
        # Test unsupported types in raise mode
        spec = YadsSpec(
            name="t",
            version="1.0.0",
            columns=[Column(name="col1", type=yads_type)],
        )

        with pytest.raises(UnsupportedFeatureError, match=f"PyArrowConverter does not support type: {re.escape(type_name)} for 'col1'."):
            PyArrowConverter(PyArrowConverterConfig(mode="raise")).convert(spec)

    def test_raise_mode_for_incompatible_decimal_precision(self):
        spec = YadsSpec(
            name="t",
            version="1.0.0",
            columns=[Column(name="d", type=Decimal(precision=39, scale=2, bits=128))],
        )

        with pytest.raises(UnsupportedFeatureError, match="precision > 38 is incompatible with Decimal\\(bits=128\\)"):
            PyArrowConverter(PyArrowConverterConfig(mode="raise")).convert(spec)

    def test_raise_mode_for_incompatible_time_bits_unit(self):
        # time32 with us -> should raise in raise mode
        spec1 = YadsSpec(
            name="t1",
            version="1.0.0",
            columns=[Column(name="t", type=Time(bits=32, unit=TimeUnit.US))],
        )
        with pytest.raises(UnsupportedFeatureError, match="time32 supports only 's' or 'ms' units"):
            PyArrowConverter(PyArrowConverterConfig(mode="raise")).convert(spec1)

        # time64 with ms -> should raise in raise mode
        spec2 = YadsSpec(
            name="t2",
            version="1.0.0",
            columns=[Column(name="t", type=Time(bits=64, unit=TimeUnit.MS))],
        )
        with pytest.raises(UnsupportedFeatureError, match="time64 supports only 'us' or 'ns' units"):
            PyArrowConverter(PyArrowConverterConfig(mode="raise")).convert(spec2)
# fmt: on


# %% PyArrowConverter column filtering and customization
class TestPyArrowConverterCustomization:
    def test_ignore_columns(self):
        """Test that ignore_columns excludes specified columns from the schema."""
        spec = YadsSpec(
            name="test",
            version="1.0.0",
            columns=[
                Column(name="id", type=Integer()),
                Column(name="name", type=String()),
                Column(name="secret", type=String()),
            ],
        )
        config = PyArrowConverterConfig(ignore_columns={"secret"})
        converter = PyArrowConverter(config)
        schema = converter.convert(spec)

        assert "id" in schema.names
        assert "name" in schema.names
        assert "secret" not in schema.names

    def test_include_columns(self):
        """Test that include_columns only includes specified columns in the schema."""
        spec = YadsSpec(
            name="test",
            version="1.0.0",
            columns=[
                Column(name="id", type=Integer()),
                Column(name="name", type=String()),
                Column(name="internal", type=String()),
            ],
        )
        config = PyArrowConverterConfig(include_columns={"id", "name"})
        converter = PyArrowConverter(config)
        schema = converter.convert(spec)

        assert "id" in schema.names
        assert "name" in schema.names
        assert "internal" not in schema.names

    def test_column_override_basic(self):
        """Test basic column override functionality."""

        def custom_name_override(field, converter):
            # Override name field to be large_string with custom metadata
            return pa.field(
                field.name,
                pa.large_string(),
                nullable=field.is_nullable,
                metadata={"custom": "true", "override": "applied"},
            )

        spec = YadsSpec(
            name="test",
            version="1.0.0",
            columns=[
                Column(name="id", type=Integer()),
                Column(name="name", type=String()),
            ],
        )
        config = PyArrowConverterConfig(column_overrides={"name": custom_name_override})
        converter = PyArrowConverter(config)
        schema = converter.convert(spec)

        # Check that override was applied
        name_field = schema.field("name")
        assert name_field.type == pa.large_string()
        assert name_field.metadata is not None
        metadata = {k.decode(): v.decode() for k, v in name_field.metadata.items()}
        assert metadata["custom"] == "true"
        assert metadata["override"] == "applied"

        # Check that other fields use default conversion
        id_field = schema.field("id")
        assert id_field.type == pa.int32()

    def test_column_override_with_complex_type(self):
        """Test column override with complex custom type."""

        def custom_metadata_override(field, converter):
            # Create a custom struct for metadata
            metadata_struct = pa.struct(
                [
                    pa.field("version", pa.string()),
                    pa.field("tags", pa.list_(pa.string())),
                    pa.field("config", pa.map_(pa.string(), pa.string())),
                ]
            )
            return pa.field(
                field.name,
                metadata_struct,
                nullable=field.is_nullable,
                metadata={"custom_struct": "true"},
            )

        spec = YadsSpec(
            name="test",
            version="1.0.0",
            columns=[
                Column(name="id", type=Integer()),
                Column(name="metadata", type=JSON()),
            ],
        )
        config = PyArrowConverterConfig(
            column_overrides={"metadata": custom_metadata_override}
        )
        converter = PyArrowConverter(config)
        schema = converter.convert(spec)

        # Check that override was applied
        metadata_field = schema.field("metadata")
        assert pa.types.is_struct(metadata_field.type)
        struct_type = metadata_field.type
        assert struct_type.num_fields == 3
        assert struct_type.field("version").type == pa.string()
        assert pa.types.is_list(struct_type.field("tags").type)
        assert pa.types.is_map(struct_type.field("config").type)

    @pytest.mark.parametrize(
        "fallback_type",
        [pa.large_string(), pa.large_binary(), pa.string(), pa.binary()],
    )
    def test_valid_fallback_types(self, fallback_type: pa.DataType):
        """Test fallback_type for unsupported types."""
        spec = YadsSpec(
            name="test",
            version="1.0.0",
            columns=[
                Column(name="id", type=Integer()),
                Column(name="geom", type=Geometry()),
            ],
        )
        config = PyArrowConverterConfig(fallback_type=fallback_type)
        converter = PyArrowConverter(config)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            schema = converter.convert(spec, mode="coerce")

        # Check fallback was applied
        geom_field = schema.field("geom")
        assert geom_field.type == fallback_type

        # Check warning was emitted
        assert len(w) == 1
        assert "geometry" in str(w[0].message)
        assert str(fallback_type) in str(w[0].message)

    def test_invalid_fallback_type_raises_error(self):
        """Test that invalid fallback_type raises UnsupportedFeatureError."""
        with pytest.raises(
            UnsupportedFeatureError,
            match="fallback_type must be one of: pa.binary\\(\\), pa.large_binary\\(\\), pa.string\\(\\), pa.large_string\\(\\)",
        ):
            PyArrowConverterConfig(fallback_type=pa.int32())

    def test_precedence_ignore_over_override(self):
        """Test that ignore_columns takes precedence over column_overrides."""

        def should_not_be_called(field, converter):
            pytest.fail("Override should not be called for ignored column")

        spec = YadsSpec(
            name="test",
            version="1.0.0",
            columns=[
                Column(name="id", type=Integer()),
                Column(name="ignored_col", type=String()),
            ],
        )
        config = PyArrowConverterConfig(
            ignore_columns={"ignored_col"},
            column_overrides={"ignored_col": should_not_be_called},
        )
        converter = PyArrowConverter(config)
        schema = converter.convert(spec)

        assert len(schema.names) == 1
        assert "id" in schema.names
        assert "ignored_col" not in schema.names

    def test_precedence_override_over_default_conversion(self):
        """Test that column_overrides takes precedence over default conversion."""

        def integer_as_string_override(field, converter):
            return pa.field(
                field.name,
                pa.string(),
                nullable=field.is_nullable,
                metadata={"converted_from": "integer"},
            )

        spec = YadsSpec(
            name="test",
            version="1.0.0",
            columns=[
                Column(name="normal_int", type=Integer()),
                Column(name="string_int", type=Integer()),
            ],
        )
        config = PyArrowConverterConfig(
            column_overrides={"string_int": integer_as_string_override}
        )
        converter = PyArrowConverter(config)
        schema = converter.convert(spec)

        # Normal conversion
        normal_field = schema.field("normal_int")
        assert normal_field.type == pa.int32()

        # Override conversion
        string_field = schema.field("string_int")
        assert string_field.type == pa.string()
        assert string_field.metadata is not None
        metadata = {k.decode(): v.decode() for k, v in string_field.metadata.items()}
        assert metadata["converted_from"] == "integer"

    def test_precedence_override_over_fallback(self):
        """Test that column_overrides takes precedence over fallback_type."""

        def custom_geometry_override(field, converter):
            return pa.field(
                field.name,
                pa.string(),
                nullable=field.is_nullable,
                metadata={"custom_geometry": "true"},
            )

        spec = YadsSpec(
            name="test",
            version="1.0.0",
            columns=[
                Column(name="fallback_geom", type=Geometry()),
                Column(name="override_geom", type=Geometry()),
            ],
        )
        config = PyArrowConverterConfig(
            fallback_type=pa.large_binary(),
            column_overrides={"override_geom": custom_geometry_override},
        )
        converter = PyArrowConverter(config)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            schema = converter.convert(spec, mode="coerce")

        # Fallback applied to fallback_geom
        fallback_field = schema.field("fallback_geom")
        assert fallback_field.type == pa.large_binary()

        # Override applied to override_geom
        override_field = schema.field("override_geom")
        assert override_field.type == pa.string()
        assert override_field.metadata is not None
        metadata = {k.decode(): v.decode() for k, v in override_field.metadata.items()}
        assert metadata["custom_geometry"] == "true"

        # Only one warning for the fallback field
        assert len(w) == 1
        assert "fallback_geom" in str(w[0].message)

    def test_field_metadata_preservation_with_fallback(self):
        """Test that field metadata is preserved when fallback is applied."""
        spec = YadsSpec(
            name="test",
            version="1.0.0",
            columns=[
                Column(
                    name="geom",
                    type=Geometry(),
                    metadata={"spatial_ref": "EPSG:4326", "precision": "high"},
                ),
            ],
        )
        config = PyArrowConverterConfig(fallback_type=pa.string())
        converter = PyArrowConverter(config)

        with warnings.catch_warnings(record=True):
            warnings.simplefilter("always")
            schema = converter.convert(spec, mode="coerce")

        geom_field = schema.field("geom")

        # Check that fallback type was applied
        assert geom_field.type == pa.string()

        # Check that field metadata was preserved during fallback
        assert geom_field.metadata is not None
        metadata = {k.decode(): v.decode() for k, v in geom_field.metadata.items()}
        assert metadata["spatial_ref"] == "EPSG:4326"
        assert metadata["precision"] == "high"

    def test_field_description_preservation_with_fallback(self):
        """Test that field description is preserved when fallback is applied."""
        spec = YadsSpec(
            name="test",
            version="1.0.0",
            columns=[
                Column(
                    name="geom",
                    type=Geometry(),
                    description="A geometry field for spatial data",
                    metadata={"spatial_ref": "EPSG:4326"},
                ),
            ],
        )
        config = PyArrowConverterConfig(fallback_type=pa.string())
        converter = PyArrowConverter(config)

        with warnings.catch_warnings(record=True):
            warnings.simplefilter("always")
            schema = converter.convert(spec, mode="coerce")

        geom_field = schema.field("geom")

        # Check that fallback type was applied
        assert geom_field.type == pa.string()

        # Check that both description and metadata were preserved during fallback
        assert geom_field.metadata is not None
        metadata = {k.decode(): v.decode() for k, v in geom_field.metadata.items()}
        assert metadata["description"] == "A geometry field for spatial data"
        assert metadata["spatial_ref"] == "EPSG:4326"

    def test_field_description_in_metadata(self):
        """Test that field descriptions are included in PyArrow field metadata."""
        spec = YadsSpec(
            name="test",
            version="1.0.0",
            columns=[
                Column(name="id", type=Integer(), description="Primary key identifier"),
                Column(
                    name="name",
                    type=String(),
                    description="User's full name",
                    metadata={"max_length": "255", "encoding": "utf-8"},
                ),
                Column(
                    name="age",
                    type=Integer(),
                    # No description
                ),
            ],
        )
        converter = PyArrowConverter()
        schema = converter.convert(spec)

        # Test field with description only
        id_field = schema.field("id")
        assert id_field.metadata is not None
        id_metadata = {k.decode(): v.decode() for k, v in id_field.metadata.items()}
        assert id_metadata["description"] == "Primary key identifier"
        assert len(id_metadata) == 1

        # Test field with both description and custom metadata
        name_field = schema.field("name")
        assert name_field.metadata is not None
        name_metadata = {k.decode(): v.decode() for k, v in name_field.metadata.items()}
        assert name_metadata["description"] == "User's full name"
        assert name_metadata["max_length"] == "255"
        assert name_metadata["encoding"] == "utf-8"
        assert len(name_metadata) == 3

        # Test field with no description or metadata
        age_field = schema.field("age")
        assert age_field.metadata is None

    def test_unknown_column_in_filters_raises_error(self):
        """Test that unknown columns in filters raise validation errors."""
        spec = YadsSpec(
            name="test",
            version="1.0.0",
            columns=[Column(name="col1", type=String())],
        )

        # Test unknown ignore_columns
        config1 = PyArrowConverterConfig(ignore_columns={"nonexistent"})
        converter1 = PyArrowConverter(config1)

        with pytest.raises(
            ConverterConfigError, match="Unknown columns in ignore_columns: nonexistent"
        ):
            converter1.convert(spec)

        # Test unknown include_columns
        config2 = PyArrowConverterConfig(include_columns={"nonexistent"})
        converter2 = PyArrowConverter(config2)

        with pytest.raises(
            ConverterConfigError, match="Unknown columns in include_columns: nonexistent"
        ):
            converter2.convert(spec)

    def test_conflicting_ignore_and_include_raises_error(self):
        """Test that overlapping ignore_columns and include_columns raises error."""
        with pytest.raises(
            ConverterConfigError, match="Columns cannot be both ignored and included"
        ):
            PyArrowConverterConfig(
                ignore_columns={"col1", "col2"}, include_columns={"col1", "col3"}
            )

    def test_column_override_preserves_nullability(self):
        """Test that column overrides preserve field nullability correctly."""

        def nullable_override(field, converter):
            return pa.field(
                field.name,
                pa.large_string(),
                nullable=field.is_nullable,  # Preserve original nullability
                metadata={"nullable_preserved": str(field.is_nullable)},
            )

        spec = YadsSpec(
            name="test",
            version="1.0.0",
            columns=[
                Column(name="nullable_col", type=String()),
                Column(
                    name="non_null_col", type=String(), constraints=[NotNullConstraint()]
                ),
            ],
        )
        config = PyArrowConverterConfig(
            column_overrides={
                "nullable_col": nullable_override,
                "non_null_col": nullable_override,
            }
        )
        converter = PyArrowConverter(config)
        schema = converter.convert(spec)

        # Check nullability preservation
        nullable_field = schema.field("nullable_col")
        assert nullable_field.nullable is True
        nullable_metadata = {
            k.decode(): v.decode() for k, v in nullable_field.metadata.items()
        }
        assert nullable_metadata["nullable_preserved"] == "True"

        non_null_field = schema.field("non_null_col")
        assert non_null_field.nullable is False
        non_null_metadata = {
            k.decode(): v.decode() for k, v in non_null_field.metadata.items()
        }
        assert non_null_metadata["nullable_preserved"] == "False"

    def test_column_override_with_original_field_access(self):
        """Test that column overrides have access to original field properties."""

        def field_inspector_override(field, converter):
            # Override that inspects the original field and creates metadata based on it
            metadata = {
                "original_type": type(field.type).__name__,
                "has_description": str(field.description is not None),
                "constraint_count": str(len(field.constraints)),
            }
            if hasattr(field.type, "length") and field.type.length is not None:
                metadata["original_length"] = str(field.type.length)

            return pa.field(
                field.name, pa.string(), nullable=field.is_nullable, metadata=metadata
            )

        spec = YadsSpec(
            name="test",
            version="1.0.0",
            columns=[
                Column(
                    name="inspected_col",
                    type=String(length=100),
                    description="A test column",
                    constraints=[NotNullConstraint()],
                ),
            ],
        )
        config = PyArrowConverterConfig(
            column_overrides={"inspected_col": field_inspector_override}
        )
        converter = PyArrowConverter(config)
        schema = converter.convert(spec)

        # Check that override had access to original field properties
        inspected_field = schema.field("inspected_col")
        metadata = {k.decode(): v.decode() for k, v in inspected_field.metadata.items()}

        assert metadata["original_type"] == "String"
        assert metadata["has_description"] == "True"
        assert metadata["constraint_count"] == "1"
        assert metadata["original_length"] == "100"


# %% Field-level fallback in complex types
class TestPyArrowConverterFieldLevelFallback:
    def test_struct_with_unsupported_field_fallback(self):
        struct_with_unsupported = Struct(
            fields=[
                Field(name="supported_field", type=String()),
                Field(name="unsupported_field", type=Geography()),
                Field(name="another_supported", type=Integer()),
            ]
        )

        spec = YadsSpec(
            name="test_struct_fallback",
            version="1.0.0",
            columns=[
                Column(name="id", type=Integer()),
                Column(name="data", type=struct_with_unsupported),
            ],
        )

        config = PyArrowConverterConfig(fallback_type=pa.string())
        converter = PyArrowConverter(config)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            schema = converter.convert(spec, mode="coerce")

        # Should have warnings for the unsupported field within the struct
        assert len(w) == 1
        assert issubclass(w[0].category, ValidationWarning)
        assert "geography" in str(w[0].message)
        assert "unsupported_field" in str(w[0].message)

        # The struct field should still be a struct, not a fallback
        data_field = schema.field("data")
        assert pa.types.is_struct(data_field.type)

        # Check individual struct fields
        struct_type = data_field.type
        assert struct_type.num_fields == 3

        # Supported fields should be converted normally
        supported_field = struct_type.field("supported_field")
        assert supported_field.type == pa.string()

        another_supported = struct_type.field("another_supported")
        assert another_supported.type == pa.int32()

        # Unsupported field should get fallback
        unsupported_field = struct_type.field("unsupported_field")
        assert unsupported_field.type == pa.string()  # fallback type

    def test_nested_struct_with_multiple_unsupported_fields(self):
        inner_struct = Struct(
            fields=[
                Field(name="inner_geom", type=Geometry()),
                Field(name="inner_string", type=String()),
            ]
        )

        outer_struct = Struct(
            fields=[
                Field(name="outer_geog", type=Geography()),
                Field(name="inner_data", type=inner_struct),
                Field(name="outer_int", type=Integer()),
            ]
        )

        spec = YadsSpec(
            name="test_nested_fallback",
            version="1.0.0",
            columns=[
                Column(name="id", type=Integer()),
                Column(name="complex_data", type=outer_struct),
            ],
        )

        config = PyArrowConverterConfig(fallback_type=pa.string())
        converter = PyArrowConverter(config)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            schema = converter.convert(spec, mode="coerce")

        # Should have warnings for both unsupported fields
        assert len(w) == 2
        assert all(issubclass(warning.category, ValidationWarning) for warning in w)

        warning_messages = [str(warning.message) for warning in w]
        assert any("geography" in msg and "outer_geog" in msg for msg in warning_messages)
        assert any("geometry" in msg and "inner_geom" in msg for msg in warning_messages)

        # The outer struct should still be a struct
        complex_field = schema.field("complex_data")
        assert pa.types.is_struct(complex_field.type)

        outer_struct_type = complex_field.type
        assert outer_struct_type.num_fields == 3

        # Check outer level fields
        outer_geog_field = outer_struct_type.field("outer_geog")
        assert outer_geog_field.type == pa.string()  # fallback

        outer_int_field = outer_struct_type.field("outer_int")
        assert outer_int_field.type == pa.int32()  # normal conversion

        # Check inner struct
        inner_data_field = outer_struct_type.field("inner_data")
        assert pa.types.is_struct(inner_data_field.type)

        inner_struct_type = inner_data_field.type
        assert inner_struct_type.num_fields == 2

        inner_geom_field = inner_struct_type.field("inner_geom")
        assert inner_geom_field.type == pa.string()  # fallback

        inner_string_field = inner_struct_type.field("inner_string")
        assert inner_string_field.type == pa.string()  # normal conversion

    def test_array_with_unsupported_element_type(self):
        array_with_unsupported = Array(element=Geography())

        spec = YadsSpec(
            name="test_array_fallback",
            version="1.0.0",
            columns=[
                Column(name="id", type=Integer()),
                Column(name="locations", type=array_with_unsupported),
            ],
        )

        config = PyArrowConverterConfig(fallback_type=pa.string())
        converter = PyArrowConverter(config)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            schema = converter.convert(spec, mode="coerce")

        # Should have warning for the unsupported element type
        assert len(w) == 1
        assert issubclass(w[0].category, ValidationWarning)
        assert "geography" in str(w[0].message)

        # The array field should still be an array, not a fallback
        locations_field = schema.field("locations")
        assert pa.types.is_list(locations_field.type)

        # The element type should be the fallback
        element_type = locations_field.type.value_type
        assert element_type == pa.string()  # fallback type

    def test_map_with_unsupported_key_or_value(self):
        # Map with unsupported key
        map_unsupported_key = Map(key=Geometry(), value=String())

        # Map with unsupported value
        map_unsupported_value = Map(key=String(), value=Geography())

        spec = YadsSpec(
            name="test_map_fallback",
            version="1.0.0",
            columns=[
                Column(name="id", type=Integer()),
                Column(name="geom_keys", type=map_unsupported_key),
                Column(name="geog_values", type=map_unsupported_value),
            ],
        )

        config = PyArrowConverterConfig(fallback_type=pa.string())
        converter = PyArrowConverter(config)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            schema = converter.convert(spec, mode="coerce")

        # Should have warnings for both unsupported types
        assert len(w) == 2
        assert all(issubclass(warning.category, ValidationWarning) for warning in w)

        warning_messages = [str(warning.message) for warning in w]
        assert any("geometry" in msg for msg in warning_messages)
        assert any("geography" in msg for msg in warning_messages)

        # Both map fields should still be maps
        geom_keys_field = schema.field("geom_keys")
        assert pa.types.is_map(geom_keys_field.type)

        geog_values_field = schema.field("geog_values")
        assert pa.types.is_map(geog_values_field.type)

        # Check key/value types
        geom_keys_map = geom_keys_field.type
        assert geom_keys_map.key_type == pa.string()  # fallback for key
        assert geom_keys_map.item_type == pa.string()  # normal conversion for value

        geog_values_map = geog_values_field.type
        assert geog_values_map.key_type == pa.string()  # normal conversion for key
        assert geog_values_map.item_type == pa.string()  # fallback for value

    def test_array_of_struct_with_unsupported_fields(self):
        struct_with_unsupported = Struct(
            fields=[
                Field(name="name", type=String()),
                Field(name="location", type=Geography()),
            ]
        )

        array_of_structs = Array(element=struct_with_unsupported)

        spec = YadsSpec(
            name="test_array_struct_fallback",
            version="1.0.0",
            columns=[
                Column(name="id", type=Integer()),
                Column(name="items", type=array_of_structs),
            ],
        )

        config = PyArrowConverterConfig(fallback_type=pa.string())
        converter = PyArrowConverter(config)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            schema = converter.convert(spec, mode="coerce")

        # Should have warning for the unsupported field within the struct
        assert len(w) == 1
        assert issubclass(w[0].category, ValidationWarning)
        assert "geography" in str(w[0].message)
        assert "location" in str(w[0].message)

        # The array field should still be an array
        items_field = schema.field("items")
        assert pa.types.is_list(items_field.type)

        # The element should be a struct
        element_type = items_field.type.value_type
        assert pa.types.is_struct(element_type)

        # Check struct fields
        assert element_type.num_fields == 2

        name_field = element_type.field("name")
        assert name_field.type == pa.string()  # normal conversion

        location_field = element_type.field("location")
        assert location_field.type == pa.string()  # fallback

    def test_complex_nested_structure_with_mixed_fallbacks(self):
        innermost_struct = Struct(
            fields=[
                Field(name="variant_field", type=Variant()),
                Field(name="normal_string", type=String()),
            ]
        )

        middle_struct = Struct(
            fields=[
                Field(name="geometry_field", type=Geometry()),
                Field(name="inner_data", type=innermost_struct),
                Field(name="normal_int", type=Integer()),
            ]
        )

        array_of_middle_structs = Array(element=middle_struct)

        map_with_unsupported = Map(key=Geography(), value=array_of_middle_structs)

        spec = YadsSpec(
            name="test_complex_nested_fallback",
            version="1.0.0",
            columns=[
                Column(name="id", type=Integer()),
                Column(name="complex_map", type=map_with_unsupported),
            ],
        )

        config = PyArrowConverterConfig(fallback_type=pa.string())
        converter = PyArrowConverter(config)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            schema = converter.convert(spec, mode="coerce")

        # Should have warnings for all unsupported types
        assert len(w) == 3
        assert all(issubclass(warning.category, ValidationWarning) for warning in w)

        warning_messages = [str(warning.message) for warning in w]
        assert any("geography" in msg for msg in warning_messages)
        assert any("geometry" in msg for msg in warning_messages)
        assert any("variant" in msg for msg in warning_messages)

        # The map field should still be a map
        complex_map_field = schema.field("complex_map")
        assert pa.types.is_map(complex_map_field.type)

        map_type = complex_map_field.type
        assert map_type.key_type == pa.string()  # fallback for Geography key

        # The value should be an array
        assert pa.types.is_list(map_type.item_type)
        array_element_type = map_type.item_type.value_type

        # The array element should be a struct
        assert pa.types.is_struct(array_element_type)
        assert array_element_type.num_fields == 3

        # Check middle struct fields
        geometry_field = array_element_type.field("geometry_field")
        assert geometry_field.type == pa.string()  # fallback

        normal_int_field = array_element_type.field("normal_int")
        assert normal_int_field.type == pa.int32()  # normal conversion

        # Check inner struct
        inner_data_field = array_element_type.field("inner_data")
        assert pa.types.is_struct(inner_data_field.type)

        inner_struct_type = inner_data_field.type
        assert inner_struct_type.num_fields == 2

        variant_field = inner_struct_type.field("variant_field")
        assert variant_field.type == pa.string()  # fallback

        normal_string_field = inner_struct_type.field("normal_string")
        assert normal_string_field.type == pa.string()  # normal conversion

    @pytest.mark.parametrize(
        "fallback_type",
        [pa.string(), pa.binary(), pa.large_string(), pa.large_binary()],
    )
    def test_fallback_type_preservation_in_nested_structures(
        self, fallback_type: pa.DataType
    ):
        struct_with_unsupported = Struct(
            fields=[
                Field(name="geom", type=Geometry()),
                Field(name="geog", type=Geography()),
                Field(name="variant", type=Variant()),
            ]
        )

        spec = YadsSpec(
            name="test_fallback_preservation",
            version="1.0.0",
            columns=[
                Column(name="id", type=Integer()),
                Column(name="data", type=struct_with_unsupported),
            ],
        )

        config = PyArrowConverterConfig(fallback_type=fallback_type)
        converter = PyArrowConverter(config)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            schema = converter.convert(spec, mode="coerce")

        # Should have warnings for all unsupported types
        assert len(w) == 3

        # The struct should still be a struct
        data_field = schema.field("data")
        assert pa.types.is_struct(data_field.type)

        struct_type = data_field.type
        assert struct_type.num_fields == 3

        # All unsupported fields should get the same fallback type
        geom_field = struct_type.field("geom")
        assert geom_field.type == fallback_type

        geog_field = struct_type.field("geog")
        assert geog_field.type == fallback_type

        variant_field = struct_type.field("variant")
        assert variant_field.type == fallback_type

    def test_field_metadata_preservation_in_nested_fallback(self):
        struct_with_metadata = Struct(
            fields=[
                Field(
                    name="geom_with_metadata",
                    type=Geometry(),
                    description="A geometry field",
                    metadata={"srid": "4326", "precision": "high"},
                ),
                Field(name="normal_field", type=String()),
            ]
        )

        spec = YadsSpec(
            name="test_nested_metadata_preservation",
            version="1.0.0",
            columns=[
                Column(name="id", type=Integer()),
                Column(name="data", type=struct_with_metadata),
            ],
        )

        config = PyArrowConverterConfig(fallback_type=pa.string())
        converter = PyArrowConverter(config)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            schema = converter.convert(spec, mode="coerce")

        # Should have warning for the unsupported field
        assert len(w) == 1

        # Check that metadata was preserved
        data_field = schema.field("data")
        struct_type = data_field.type

        geom_field = struct_type.field("geom_with_metadata")
        assert geom_field.type == pa.string()  # fallback applied

        # Check that metadata was preserved
        assert geom_field.metadata is not None
        metadata = {k.decode(): v.decode() for k, v in geom_field.metadata.items()}
        assert metadata["description"] == "A geometry field"
        assert metadata["srid"] == "4326"
        assert metadata["precision"] == "high"

        # Normal field should work as expected
        normal_field = struct_type.field("normal_field")
        assert normal_field.type == pa.string()

    def test_raise_mode_still_raises_for_nested_unsupported_types(self):
        struct_with_unsupported = Struct(
            fields=[
                Field(name="geom", type=Geometry()),
                Field(name="normal_field", type=String()),
            ]
        )

        spec = YadsSpec(
            name="test_raise_mode_nested",
            version="1.0.0",
            columns=[
                Column(name="id", type=Integer()),
                Column(name="data", type=struct_with_unsupported),
            ],
        )

        config = PyArrowConverterConfig(mode="raise")
        converter = PyArrowConverter(config)

        with pytest.raises(
            UnsupportedFeatureError,
            match="PyArrowConverter does not support type: geometry for 'geom'.",
        ):
            converter.convert(spec)



"""Custom yads exceptions and shared warning utilities."""

from __future__ import annotations

import warnings


class YadsError(Exception):
    """Base exception for all yads-related errors.

    This is the root exception that all other yads exceptions inherit from.
    It provides enhanced error reporting with suggestions for resolution.

    Attributes:
        suggestions: List of suggested fixes or actions.

    Example:
        >>> raise YadsError(
        ...     "Something went wrong with field 'user_id' at line 42",
        ...     suggestions=["Check the field name", "Verify the type definition"]
        ... )
    """

    def __init__(
        self,
        message: str,
        suggestions: list[str] | None = None,
    ):
        """Initialize a YadsError.

        Args:
            message: The error message.
            suggestions: Optional list of suggestions to fix the error.
        """
        super().__init__(message)
        self.suggestions = suggestions or []

    def __str__(self) -> str:
        result = super().__str__()

        if self.suggestions:
            suggestions_text = "; ".join(self.suggestions)
            result += f" | {suggestions_text}"

        return result


class YadsValidationError(YadsError):
    """Base for all validation-related errors."""


# Spec Exceptions
class SpecError(YadsValidationError):
    """Spec definition and validation errors."""


class SpecParsingError(SpecError):
    """Errors during spec parsing from YAML/JSON."""


class SpecValidationError(SpecError):
    """Spec consistency and integrity validation errors."""


class SpecSerializationError(SpecError):
    """Errors encountered while serializing specs to dictionaries."""


# Type System Exceptions
class TypeDefinitionError(YadsValidationError):
    """Invalid type definitions and parameters."""


class UnknownTypeError(TypeDefinitionError):
    """Unknown or unsupported type name."""


# Constraint Exceptions
class ConstraintError(YadsValidationError):
    """Constraint definition and validation errors."""


class UnknownConstraintError(ConstraintError):
    """Unknown constraint type."""


class InvalidConstraintError(ConstraintError):
    """Invalid constraint parameters or configuration."""


class ConstraintConflictError(ConstraintError):
    """Conflicting constraints."""


# Cross-cutting Exceptions (used across multiple domains)
class ConfigError(YadsError):
    """Base for configuration-related errors."""


class UnsupportedFeatureError(YadsError):
    """Feature not supported by target system/component."""


# Dependency Exceptions
class DependencyError(YadsError):
    """Base for optional dependency errors."""


class MissingDependencyError(DependencyError):
    """Required optional dependency is not installed."""


class DependencyVersionError(DependencyError):
    """Installed dependency version does not meet requirements."""


# Converter Exceptions
class ConverterError(YadsError):
    """Base for converter-related errors."""


class ConverterConfigError(ConfigError):
    """Errors during converter configuration."""


class ConversionError(ConverterError):
    """Errors during conversion process."""


# Loader Exceptions
class LoaderError(YadsError):
    """Base for loader-related errors."""


class LoaderConfigError(ConfigError):
    """Errors during loader configuration."""


# Registry Exceptions
class RegistryError(YadsError):
    """Base exception for registry operations."""


class RegistryConnectionError(RegistryError):
    """Failed to connect or validate registry base path."""


class SpecNotFoundError(RegistryError):
    """Spec name or version not found in registry."""


class InvalidSpecNameError(RegistryError):
    """Spec name contains invalid characters."""


# Validator Exceptions
class AstValidationError(YadsValidationError):
    """SQL AST validation rule processing errors."""


# Shared warnings
class ValidationWarning(UserWarning):
    """Warning emitted when validation rules fail in converters or validators."""


class DuplicateSpecWarning(UserWarning):
    """Warning emitted when registering a spec with identical content to latest version."""


def validation_warning(message: str, *, filename: str, module: str | None = None) -> None:
    """Emit a categorized validation warning with a concise origin.

    Args:
        message: Human-readable warning message.
        filename: Logical filename/module label to display as the source.
        module: Module name override. Defaults to the caller's module if not provided.
    """
    warnings.warn_explicit(
        message=message,
        category=ValidationWarning,
        filename=filename,
        lineno=1,
        module=module or __name__,
    )



"""Executable example for converting a yads spec into a Pydantic model."""

from __future__ import annotations

from ..base import ExampleBlockRequest, ExampleDefinition

import warnings

warnings.simplefilter("ignore")


def _pydantic_model_example() -> None:
    import yads
    from yads.converters import PydanticConverter, PydanticConverterConfig
    import json

    spec = yads.from_yaml("docs/src/specs/submissions.yaml")

    converter = PydanticConverter(PydanticConverterConfig(mode="coerce"))
    Submission = converter.convert(spec)
    print(json.dumps(Submission.model_json_schema(), indent=2))


EXAMPLE = ExampleDefinition(
    example_id="pydantic-converter-basic",
    blocks=(
        ExampleBlockRequest(
            slug="code",
            language="python",
            source="callable",
            callable=_pydantic_model_example,
        ),
        ExampleBlockRequest(
            slug="output",
            language="text",
            source="stdout",
            callable=_pydantic_model_example,
        ),
    ),
)


if __name__ == "__main__":
    _pydantic_model_example()



#!/usr/bin/env python3
"""Spark integration test for SQL and PySpark converters.

This script validates that:
1. SparkSqlConverter can generate valid DDL for Spark + Iceberg
2. PySparkConverter can generate valid StructType schemas
"""

import sys
from pathlib import Path

import yads
from yads.converters.sql import SparkSqlConverter
from yads.converters import PySparkConverter


def test_spark_sql_converter():
    """Test Spark SQL converter with iceberg_spec fixture."""
    print("=" * 60)
    print("Spark SQL Converter Integration Test")
    print("=" * 60)

    from pyspark.sql import SparkSession

    fixture_path = (
        Path(__file__).parents[3] / "tests/fixtures/spec/valid/iceberg_spec.yaml"
    )
    print(f"\n▶ Loading spec from: {fixture_path}")

    spec = yads.from_yaml(str(fixture_path))
    print(f"✓ Loaded spec: {spec.name} (version {spec.version})")

    print("\n▶ Generating DDL with SparkSqlConverter...")
    from yads.converters.sql.ast_converter import SqlglotConverterConfig

    ast_config = SqlglotConverterConfig(ignore_catalog=True, ignore_database=True)
    converter = SparkSqlConverter(mode="coerce", ast_config=ast_config)
    ddl = converter.convert(spec, pretty=True)

    print("\n--- Generated DDL ---")
    print(ddl)
    print("--- End DDL ---\n")

    print("▶ Creating SparkSession with Iceberg Hadoop catalog...")
    spark = (
        SparkSession.builder.appName("YadsIntegrationTest")
        .master("local[*]")
        .config(
            "spark.sql.extensions",
            "org.apache.iceberg.spark.extensions.IcebergSparkSessionExtensions",
        )
        .config(
            "spark.sql.catalog.spark_catalog", "org.apache.iceberg.spark.SparkCatalog"
        )
        .config("spark.sql.catalog.spark_catalog.type", "hadoop")
        .config("spark.sql.catalog.spark_catalog.warehouse", "/tmp/iceberg-warehouse")
        .config("spark.sql.warehouse.dir", "/tmp/spark-warehouse")
        .config("spark.sql.catalogImplementation", "in-memory")
        .getOrCreate()
    )

    print("▶ Executing DDL...")
    try:
        spark.sql(ddl)
        print("✓ DDL executed successfully!")
    except Exception as e:
        print(f"✗ DDL execution failed: {e}")
        raise
    finally:
        spark.stop()

    print("\n" + "=" * 60)
    print("✓ Spark SQL integration test PASSED")
    print("=" * 60)


def test_pyspark_converter():
    """Test PySpark converter with iceberg_spec fixture."""
    print("\n" + "=" * 60)
    print("PySpark Converter Integration Test")
    print("=" * 60)

    from pyspark.sql import SparkSession

    fixture_path = (
        Path(__file__).parents[3] / "tests/fixtures/spec/valid/iceberg_spec.yaml"
    )
    print(f"\n▶ Loading spec from: {fixture_path}")

    spec = yads.from_yaml(str(fixture_path))
    print(f"✓ Loaded spec: {spec.name} (version {spec.version})")

    print("\n▶ Generating schema with PySparkConverter...")
    from yads.converters import PySparkConverterConfig

    config = PySparkConverterConfig(mode="coerce")
    converter = PySparkConverter(config=config)
    schema = converter.convert(spec)

    print("\n--- Generated Schema ---")
    print(schema.simpleString())
    print("--- End Schema ---\n")

    print("▶ Creating SparkSession...")
    spark = (
        SparkSession.builder.appName("YadsSchemaTest").master("local[*]").getOrCreate()
    )

    print("▶ Creating DataFrame with generated schema...")
    try:
        df = spark.createDataFrame([], schema)
        print("✓ DataFrame created successfully!")
        print(f"  Schema columns: {df.columns}")
    except Exception as e:
        print(f"✗ DataFrame creation failed: {e}")
        raise
    finally:
        spark.stop()

    print("\n" + "=" * 60)
    print("✓ PySpark converter integration test PASSED")
    print("=" * 60)


if __name__ == "__main__":
    exit_code = 0

    try:
        test_spark_sql_converter()
    except Exception as e:
        print(f"\n✗ Spark SQL test FAILED: {e}", file=sys.stderr)
        import traceback

        traceback.print_exc()
        exit_code = 1

    try:
        test_pyspark_converter()
    except Exception as e:
        print(f"\n✗ PySpark converter test FAILED: {e}", file=sys.stderr)
        import traceback

        traceback.print_exc()
        exit_code = 1

    sys.exit(exit_code)



"""AST validation engine for dialect compatibility."""

from __future__ import annotations

# pyright: reportUnknownArgumentType=none, reportUnknownMemberType=none
# pyright: reportUnknownVariableType=none

from typing import TYPE_CHECKING, Literal, cast

from ...._dependencies import ensure_dependency
from ....exceptions import AstValidationError, validation_warning
from .ast_validation_rules import AstValidationRule

ensure_dependency("sqlglot")

from sqlglot.expressions import Create  # noqa: E402

if TYPE_CHECKING:
    from sqlglot import exp


class AstValidator:
    """Apply a list of `AstValidationRule` instances to a sqlglot AST.

    AstValidator applies a set of validation/adjustment rules to a sqlglot
    AST in two modes:
    - "raise": collect all errors and raise
    - "coerce": apply adjustments and emit warnings

    The validator traverses the AST recursively, applying each rule to every
    node.

    Args:
        rules: List of `AstValidationRule` instances to apply during validation.

    Example:
        >>> from yads.converters.sql import AstValidator, DisallowFixedLengthString
        >>>
        >>> # Create validator with built-in rules
        >>> rules = [DisallowFixedLengthString()]
        >>> validator = AstValidator(rules=rules)
        >>>
        >>> # Apply validation in different modes
        >>> try:
        ...     ast = validator.validate(ast, mode="raise")
        ... except AstValidationError as e:
        ...     print(f"Validation failed: {e}")
        >>>
        >>> # Auto-fix with warnings
        >>> ast = validator.validate(ast, mode="coerce")
    """

    def __init__(self, rules: list[AstValidationRule]):
        self.rules = rules

    def validate(self, ast: exp.Create, mode: Literal["raise", "coerce"]) -> exp.Create:
        errors: list[str] = []

        def transformer(node: exp.Expression) -> exp.Expression:
            for rule in self.rules:
                error = rule.validate(node)
                if not error:
                    continue
                match mode:
                    case "raise":
                        errors.append(f"{error}")
                    case "coerce":
                        validation_warning(
                            message=f"{error} {rule.adjustment_description}",
                            filename="yads.converters.sql.validators",
                            module=__name__,
                        )
                        node = rule.adjust(node)
                    case _:
                        raise AstValidationError(f"Invalid mode: {mode}.")
            return node

        processed_ast = ast.transform(transformer, copy=False)

        if errors:
            error_summary = "\n".join(f"- {e}" for e in errors)
            raise AstValidationError(
                "Validation for the target dialect failed with the following errors:\n"
                f"{error_summary}"
            )

        return cast(Create, processed_ast)



from __future__ import annotations

import pytest
from sqlglot import parse_one
from sqlglot.expressions import ColumnDef, DataType

from yads.converters.sql.validators.ast_validation_rules import (
    DisallowType,
    DisallowUserDefinedType,
    DisallowFixedLengthString,
    DisallowFixedLengthBinary,
    DisallowNegativeScaleDecimal,
    DisallowParameterizedGeometry,
    DisallowColumnConstraintGeneratedIdentity,
    DisallowTableConstraintPrimaryKeyNullsFirst,
)


# %% DisallowType
class TestDisallowType:
    @pytest.fixture
    def rule(self) -> DisallowType:
        return DisallowType(disallow_type=DataType.Type.JSON)

    @pytest.mark.parametrize(
        "sql, expected",
        [
            (
                "JSON",
                "Data type 'JSON' is not supported for column 'col'.",
            ),
            ("INT", None),
            ("STRING", None),
        ],
    )
    def test_validate_disallowed_type_json(
        self, rule: DisallowType, sql: str, expected: str | None
    ):
        ast = parse_one(f"CREATE TABLE t (col {sql})")
        assert ast
        column_def = ast.find(ColumnDef)
        assert column_def
        data_type = column_def.find(DataType)
        assert data_type

        assert rule.validate(data_type) == expected

    def test_adjust_replaces_disallowed_type_with_default(self, rule: DisallowType):
        ast = parse_one("CREATE TABLE t (col JSON)")
        assert ast
        data_type = ast.find(DataType)
        assert data_type

        adjusted_node = rule.adjust(data_type)

        assert isinstance(adjusted_node, DataType)
        assert adjusted_node.this == DataType.Type.TEXT
        assert not adjusted_node.expressions

    def test_adjustment_description(self, rule: DisallowType):
        assert (
            rule.adjustment_description == "The data type will be replaced with 'TEXT'."
        )

    def test_adjust_with_custom_fallback(self):
        rule = DisallowType(
            disallow_type=DataType.Type.JSON, fallback_type=DataType.Type.VARCHAR
        )
        ast = parse_one("CREATE TABLE t (col JSON)")
        assert ast
        data_type = ast.find(DataType)
        assert data_type

        adjusted_node = rule.adjust(data_type)

        assert isinstance(adjusted_node, DataType)
        assert adjusted_node.this == DataType.Type.VARCHAR
        assert not adjusted_node.expressions
        assert (
            rule.adjustment_description
            == "The data type will be replaced with 'VARCHAR'."
        )


# %% DisallowUserDefinedType
class TestDisallowUserDefinedType:
    @pytest.mark.parametrize(
        "rule, sql, expected",
        [
            (
                DisallowUserDefinedType(disallow_type="VOID"),
                "VOID",
                "Data type 'VOID' is not supported for column 'col'.",
            ),
            (DisallowUserDefinedType(disallow_type="VOID"), "GEOMETRY", None),
            (DisallowUserDefinedType(disallow_type="VOID"), "TEXT", None),
        ],
    )
    def test_validate_user_defined_type(
        self, rule: DisallowUserDefinedType, sql: str, expected: str | None
    ):
        ast = parse_one(f"CREATE TABLE t (col {sql})")
        assert ast
        column_def = ast.find(ColumnDef)
        assert column_def
        data_type = column_def.find(DataType)
        assert data_type

        assert rule.validate(data_type) == expected

    def test_adjust_replaces_user_defined_type_with_fallback(self):
        rule = DisallowUserDefinedType(
            disallow_type="VOID", fallback_type=DataType.Type.TEXT
        )
        ast = parse_one("CREATE TABLE t (col VOID)")
        assert ast
        data_type = ast.find(DataType)
        assert data_type

        adjusted = rule.adjust(data_type)

        assert isinstance(adjusted, DataType)
        assert adjusted.this == DataType.Type.TEXT
        assert not adjusted.expressions

    def test_adjustment_description(self):
        rule = DisallowUserDefinedType(
            disallow_type="VOID", fallback_type=DataType.Type.TEXT
        )
        assert (
            rule.adjustment_description == "The data type will be replaced with 'TEXT'."
        )


# %% DisallowFixedLengthString
class TestDisallowFixedLengthString:
    @pytest.fixture
    def rule(self) -> DisallowFixedLengthString:
        return DisallowFixedLengthString()

    @pytest.mark.parametrize(
        "sql, expected",
        [
            ("VARCHAR(50)", "Fixed-length strings are not supported for column 'col'."),
            (
                "CHAR(10)",
                "Fixed-length strings are not supported for column 'col'.",
            ),
            (
                "STRING",
                None,
            ),
            ("INT", None),
        ],
    )
    def test_validate_fixed_length_strings(
        self, rule: DisallowFixedLengthString, sql: str, expected: str | None
    ):
        ast = parse_one(f"CREATE TABLE t (col {sql})")
        assert ast
        column_def = ast.find(ColumnDef)
        assert column_def
        data_type = column_def.find(DataType)
        assert data_type

        assert rule.validate(data_type) == expected

    def test_adjust_removes_length_and_normalizes_type(
        self, rule: DisallowFixedLengthString
    ):
        sql = "CREATE TABLE t (col VARCHAR(50))"
        ast = parse_one(sql)
        assert ast
        data_type = ast.find(DataType)
        assert data_type

        adjusted_node = rule.adjust(data_type)

        assert isinstance(adjusted_node, DataType)
        assert adjusted_node.this == DataType.Type.VARCHAR
        assert not adjusted_node.expressions

    def test_adjustment_description(self, rule: DisallowFixedLengthString):
        assert rule.adjustment_description == "The length parameter will be removed."


# %% DisallowFixedLengthBinary
class TestDisallowFixedLengthBinary:
    @pytest.fixture
    def rule(self) -> DisallowFixedLengthBinary:
        return DisallowFixedLengthBinary()

    @pytest.mark.parametrize(
        "sql, expected",
        [
            ("BINARY(50)", "Fixed-length binary is not supported for column 'col'."),
            (
                "BINARY",
                None,
            ),
            ("INT", None),
        ],
    )
    def test_validate_fixed_length_binary(
        self, rule: DisallowFixedLengthBinary, sql: str, expected: str | None
    ):
        ast = parse_one(f"CREATE TABLE t (col {sql})")
        assert ast
        column_def = ast.find(ColumnDef)
        assert column_def
        data_type = column_def.find(DataType)
        assert data_type

        assert rule.validate(data_type) == expected

    def test_adjust_removes_length_and_normalizes_type(
        self, rule: DisallowFixedLengthBinary
    ):
        sql = "CREATE TABLE t (col BINARY(50))"
        ast = parse_one(sql)
        assert ast
        data_type = ast.find(DataType)
        assert data_type

        adjusted_node = rule.adjust(data_type)

        assert isinstance(adjusted_node, DataType)
        assert adjusted_node.this == DataType.Type.BINARY
        assert not adjusted_node.expressions

    def test_adjustment_description(self, rule: DisallowFixedLengthBinary):
        assert rule.adjustment_description == "The length parameter will be removed."


# %% DisallowParameterizedGeometry
class TestDisallowParameterizedGeometry:
    @pytest.fixture
    def rule(self) -> DisallowParameterizedGeometry:
        return DisallowParameterizedGeometry()

    @pytest.mark.parametrize(
        "sql, expected",
        [
            (
                "GEOMETRY(4326)",
                "Parameterized 'GEOMETRY' is not supported for column 'col'.",
            ),
            ("GEOMETRY", None),
            ("INT", None),
        ],
    )
    def test_validate_parameterized_geometry(
        self, rule: DisallowParameterizedGeometry, sql: str, expected: str | None
    ):
        ast = parse_one(f"CREATE TABLE t (col {sql})")
        assert ast
        column_def = ast.find(ColumnDef)
        assert column_def
        data_type = column_def.find(DataType)
        assert data_type

        assert rule.validate(data_type) == expected

    def test_adjust_removes_geometry_parameters(
        self, rule: DisallowParameterizedGeometry
    ):
        ast = parse_one("CREATE TABLE t (col GEOMETRY(4326))")
        assert ast
        data_type = ast.find(DataType)
        assert data_type

        adjusted = rule.adjust(data_type)

        assert isinstance(adjusted, DataType)
        assert adjusted.this == DataType.Type.GEOMETRY
        assert not adjusted.expressions

    def test_adjustment_description(self, rule: DisallowParameterizedGeometry):
        assert rule.adjustment_description == "The parameters will be removed."


# %% DisallowNegativeScaleDecimal
class TestDisallowNegativeScaleDecimal:
    @pytest.fixture
    def rule(self) -> DisallowNegativeScaleDecimal:
        return DisallowNegativeScaleDecimal()

    @pytest.fixture
    def decimal_datatype_in_create(self):
        # Build a full CREATE TABLE AST and return the nested DECIMAL DataType node
        # to ensure ancestor context (column name) is available during validation.
        from sqlglot.expressions import (
            Create,
            Schema,
            Table,
            Identifier,
            ColumnDef,
            DataType,
            DataTypeParam,
            Literal,
        )

        def _build(precision: int, scale: int) -> DataType:
            dtype = DataType(
                this=DataType.Type.DECIMAL,
                expressions=[
                    DataTypeParam(this=Literal(this=precision)),
                    DataTypeParam(this=Literal(this=scale)),
                ],
            )
            coldef = ColumnDef(this=Identifier(this="col"), kind=dtype)
            _ = Create(
                this=Schema(this=Table(this=Identifier(this="t")), expressions=[coldef]),
                kind="TABLE",
            )
            # Returning dtype keeps it attached in the AST so ancestors are discoverable
            return dtype

        return _build

    def test_validate_detects_negative_scale_decimal(
        self, rule: DisallowNegativeScaleDecimal, decimal_datatype_in_create
    ):
        data_type = decimal_datatype_in_create(10, -2)
        result = rule.validate(data_type)
        assert (
            result
            == "Negative scale decimals are not supported for column 'col'. Found DECIMAL(10, -2)."
        )

    def test_validate_ignores_positive_scale_decimal(
        self, rule: DisallowNegativeScaleDecimal, decimal_datatype_in_create
    ):
        data_type = decimal_datatype_in_create(10, 2)
        result = rule.validate(data_type)
        assert result is None

    def test_validate_ignores_zero_scale_decimal(
        self, rule: DisallowNegativeScaleDecimal, decimal_datatype_in_create
    ):
        data_type = decimal_datatype_in_create(10, 0)
        result = rule.validate(data_type)
        assert result is None

    def test_validate_ignores_non_decimal_types(self, rule: DisallowNegativeScaleDecimal):
        from sqlglot.expressions import DataType

        # Test INT - should not trigger validation
        data_type = DataType(this=DataType.Type.INT, expressions=[])

        result = rule.validate(data_type)
        assert result is None

    def test_adjust_coerces_negative_scale_to_positive_scale(
        self, rule: DisallowNegativeScaleDecimal, decimal_datatype_in_create
    ):
        # Test DECIMAL(10, -2) -> DECIMAL(12, 0)
        data_type = decimal_datatype_in_create(10, -2)
        adjusted_node = rule.adjust(data_type)

        assert isinstance(adjusted_node, DataType)
        assert adjusted_node.this == DataType.Type.DECIMAL
        assert len(adjusted_node.expressions) == 2
        assert adjusted_node.expressions[0].this.this == 12  # 10 + abs(-2)
        assert adjusted_node.expressions[1].this.this == 0  # scale set to 0

    def test_adjust_coerces_large_negative_scale(
        self, rule: DisallowNegativeScaleDecimal, decimal_datatype_in_create
    ):
        # Test DECIMAL(10, -6) -> DECIMAL(16, 0)
        data_type = decimal_datatype_in_create(10, -6)
        adjusted_node = rule.adjust(data_type)

        assert isinstance(adjusted_node, DataType)
        assert adjusted_node.this == DataType.Type.DECIMAL
        assert len(adjusted_node.expressions) == 2
        assert adjusted_node.expressions[0].this.this == 16  # 10 + abs(-6)
        assert adjusted_node.expressions[1].this.this == 0  # scale set to 0

    def test_adjust_does_not_modify_valid_decimal(
        self, rule: DisallowNegativeScaleDecimal, decimal_datatype_in_create
    ):
        # Test that valid decimals are not modified
        data_type = decimal_datatype_in_create(10, 2)
        adjusted_node = rule.adjust(data_type)

        # Should return the same node unchanged
        assert adjusted_node is data_type

    def test_adjustment_description(self, rule: DisallowNegativeScaleDecimal):
        assert (
            rule.adjustment_description
            == "The precision will be increased by the absolute value of the negative scale, and the scale will be set to 0."
        )


# %% DisallowColumnConstraintGeneratedIdentity
class TestDisallowColumnConstraintGeneratedIdentity:
    @pytest.fixture
    def rule(self) -> DisallowColumnConstraintGeneratedIdentity:
        return DisallowColumnConstraintGeneratedIdentity()

    def test_validate_detects_identity_constraint(
        self, rule: DisallowColumnConstraintGeneratedIdentity
    ):
        sql = """
        CREATE TABLE t (
          c1 INT GENERATED ALWAYS AS IDENTITY(1, 1),
          c2 TEXT
        )
        """
        ast = parse_one(sql)
        coldef = ast.find(ColumnDef)
        assert coldef
        assert (
            rule.validate(coldef)
            == "GENERATED ALWAYS AS IDENTITY is not supported for column 'c1'."
        )

    def test_adjust_removes_identity_constraint(
        self, rule: DisallowColumnConstraintGeneratedIdentity
    ):
        sql = "CREATE TABLE t (c1 INT GENERATED ALWAYS AS IDENTITY(1, 1))"
        ast = parse_one(sql)
        coldef = ast.find(ColumnDef)
        assert coldef
        adjusted = rule.adjust(coldef)
        assert isinstance(adjusted, DataType) or isinstance(adjusted, ColumnDef)
        # Ensure no GeneratedAsIdentityColumnConstraint remains
        if isinstance(adjusted, ColumnDef):
            constraints = adjusted.args.get("constraints") or []
            assert all(
                type(c.kind).__name__ != "GeneratedAsIdentityColumnConstraint"
                for c in constraints
            )


# %% DisallowTableConstraintPrimaryKeyNullsFirst
class TestDisallowTableConstraintPrimaryKeyNullsFirst:
    @pytest.fixture
    def rule(self) -> DisallowTableConstraintPrimaryKeyNullsFirst:
        return DisallowTableConstraintPrimaryKeyNullsFirst()

    def test_validate_detects_nulls_first_in_pk(
        self, rule: DisallowTableConstraintPrimaryKeyNullsFirst
    ):
        sql = "CREATE TABLE t (c1 INT, CONSTRAINT pk PRIMARY KEY (c1 NULLS FIRST))"
        ast = parse_one(sql)
        from sqlglot.expressions import PrimaryKey

        node = ast.find(PrimaryKey)
        assert node is not None
        assert (
            rule.validate(node)
            == "NULLS FIRST is not supported in PRIMARY KEY constraints."
        )

    def test_adjust_removes_nulls_first_in_pk(
        self, rule: DisallowTableConstraintPrimaryKeyNullsFirst
    ):
        sql = "CREATE TABLE t (c1 INT, CONSTRAINT pk PRIMARY KEY (c1 NULLS FIRST))"
        ast = parse_one(sql)
        from sqlglot.expressions import PrimaryKey, Ordered

        node = ast.find(PrimaryKey)
        assert node is not None
        adjusted = rule.adjust(node)
        for expr in adjusted.expressions:
            if isinstance(expr, Ordered):
                assert not expr.args.get("nulls_first")



import pytest
import warnings
from sqlglot import exp

from yads.spec import YadsSpec, Column, Field
from yads.converters.sql import DuckdbSqlConverter, SqlglotConverterConfig
from yads.exceptions import AstValidationError
from yads.loaders import from_yaml_string, from_yaml_path
from yads.exceptions import ValidationWarning
from yads.types import (
    YadsType,
    String,
    Integer,
    Float,
    Decimal,
    Boolean,
    Binary,
    Date,
    TimeUnit,
    Time,
    Timestamp,
    TimestampTZ,
    TimestampLTZ,
    TimestampNTZ,
    Duration,
    IntervalTimeUnit,
    Interval,
    Array,
    Struct,
    Map,
    JSON,
    Geometry,
    Geography,
    UUID,
    Void,
    Variant,
)


# %% Types
class TestDuckdbSqlConverterTypes:
    @pytest.mark.parametrize(
        "yads_type, expected_sql, expected_warning",
        [
            (String(), "TEXT", None),
            (String(length=255), "TEXT(255)", None),
            (Integer(bits=8), "TINYINT", None),
            (Integer(bits=16), "SMALLINT", None),
            (Integer(bits=32), "INT", None),
            (Integer(bits=64), "BIGINT", None),
            (Integer(bits=8, signed=False), "UTINYINT", None),
            (
                Integer(bits=16, signed=False),
                "USMALLINT",
                None,
            ),
            (Integer(bits=32, signed=False), "UINTEGER", None),
            (
                Integer(bits=64, signed=False),
                "UBIGINT",
                None,
            ),
            (
                Float(bits=16),
                "REAL",
                "SqlglotConverter does not support half-precision Float (bits=16).",
            ),
            (Float(bits=32), "REAL", None),
            (Float(bits=64), "DOUBLE", None),
            (Decimal(), "DECIMAL", None),
            (Decimal(precision=10, scale=2), "DECIMAL(10, 2)", None),
            (
                Decimal(precision=10, scale=-2),
                "DECIMAL(12)",
                "The precision will be increased by the absolute value of the negative scale, and the scale will be set to 0.",
            ),
            (Decimal(precision=10, scale=2, bits=128), "DECIMAL(10, 2)", None),
            (Boolean(), "BOOLEAN", None),
            (Binary(), "BLOB", None),
            (Binary(length=8), "BLOB", "The length parameter will be removed."),
            (Date(), "DATE", None),
            (Date(bits=32), "DATE", None),
            (Date(bits=64), "DATE", None),
            (Time(), "TIME", None),
            (Time(unit=TimeUnit.S), "TIME", None),
            (Time(unit=TimeUnit.MS), "TIME", None),
            (Time(unit=TimeUnit.US), "TIME", None),
            (Time(unit=TimeUnit.NS), "TIME", None),
            (Time(bits=32), "TIME", None),
            (Time(bits=64), "TIME", None),
            (Timestamp(), "TIMESTAMP", None),
            (Timestamp(unit=TimeUnit.S), "TIMESTAMP", None),
            (Timestamp(unit=TimeUnit.MS), "TIMESTAMP", None),
            (Timestamp(unit=TimeUnit.US), "TIMESTAMP", None),
            (Timestamp(unit=TimeUnit.NS), "TIMESTAMP", None),
            (TimestampTZ(), "TIMESTAMPTZ", None),
            (TimestampTZ(unit=TimeUnit.S), "TIMESTAMPTZ", None),
            (TimestampTZ(unit=TimeUnit.MS), "TIMESTAMPTZ", None),
            (TimestampTZ(unit=TimeUnit.US), "TIMESTAMPTZ", None),
            (TimestampTZ(unit=TimeUnit.NS), "TIMESTAMPTZ", None),
            (TimestampTZ(tz="UTC"), "TIMESTAMPTZ", None),
            (
                TimestampLTZ(),
                "TIMESTAMPTZ",
                "Data type 'TIMESTAMPLTZ' is not supported for column 'col1'.",
            ),
            (
                TimestampLTZ(unit=TimeUnit.S),
                "TIMESTAMPTZ",
                "Data type 'TIMESTAMPLTZ' is not supported for column 'col1'.",
            ),
            (
                TimestampLTZ(unit=TimeUnit.MS),
                "TIMESTAMPTZ",
                "Data type 'TIMESTAMPLTZ' is not supported for column 'col1'.",
            ),
            (
                TimestampLTZ(unit=TimeUnit.US),
                "TIMESTAMPTZ",
                "Data type 'TIMESTAMPLTZ' is not supported for column 'col1'.",
            ),
            (
                TimestampLTZ(unit=TimeUnit.NS),
                "TIMESTAMPTZ",
                "Data type 'TIMESTAMPLTZ' is not supported for column 'col1'.",
            ),
            (
                TimestampNTZ(),
                "TIMESTAMP",
                None,
            ),  # Default TIMESTAMP in Duckdb is timezone unaware
            (TimestampNTZ(unit=TimeUnit.S), "TIMESTAMP", None),
            (TimestampNTZ(unit=TimeUnit.MS), "TIMESTAMP", None),
            (TimestampNTZ(unit=TimeUnit.US), "TIMESTAMP", None),
            (TimestampNTZ(unit=TimeUnit.NS), "TIMESTAMP", None),
            (
                Duration(),
                "TEXT",
                "SqlglotConverter does not support type: duration(unit=ns)",
            ),
            (Interval(interval_start=IntervalTimeUnit.DAY), "INTERVAL DAY", None),
            (
                Interval(
                    interval_start=IntervalTimeUnit.DAY, interval_end=IntervalTimeUnit.DAY
                ),
                "INTERVAL DAY",
                None,
            ),
            (
                Interval(
                    interval_start=IntervalTimeUnit.YEAR,
                    interval_end=IntervalTimeUnit.MONTH,
                ),
                "INTERVAL YEAR TO MONTH",
                None,
            ),
            (Array(element=Integer()), "INT[]", None),
            (Array(element=String(), size=2), "TEXT[]", None),
            (
                Struct(
                    fields=[
                        Field(name="nested_int", type=Integer()),
                        Field(name="nested_string", type=String()),
                    ]
                ),
                "STRUCT(nested_int INT, nested_string TEXT)",
                None,
            ),
            (
                Struct(
                    fields=[
                        Field(name="nested_int", type=Integer()),
                        Field(
                            name="nested_struct",
                            type=Struct(
                                fields=[Field(name="nested_string", type=String())]
                            ),
                        ),
                    ]
                ),
                "STRUCT(nested_int INT, nested_struct STRUCT(nested_string TEXT))",
                None,
            ),
            (Map(key=String(), value=Integer()), "MAP(TEXT, INT)", None),
            (
                Map(key=String(), value=Array(element=String())),
                "MAP(TEXT, TEXT[])",
                None,
            ),
            (
                Map(key=String(), value=Integer(), keys_sorted=True),
                "MAP(TEXT, INT)",
                None,
            ),
            (JSON(), "JSON", None),
            (Geometry(), "GEOMETRY", None),
            (
                Geometry(srid=4326),
                "GEOMETRY",
                "Parameterized 'GEOMETRY' is not supported for column 'col1'.",
            ),
            (
                Geography(),
                "TEXT",
                "Data type 'GEOGRAPHY' is not supported for column 'col1'.",
            ),
            (
                Geography(srid=4326),
                "TEXT",
                "Data type 'GEOGRAPHY' is not supported for column 'col1'.",
            ),
            (UUID(), "UUID", None),
            (Void(), "TEXT", "Data type 'VOID' is not supported for column 'col1'."),
            (
                Variant(),
                "TEXT",
                "Data type 'VARIANT' is not supported for column 'col1'.",
            ),
        ],
    )
    def test_convert_type(
        self, yads_type: YadsType, expected_sql: str, expected_warning: str | None
    ):
        spec = YadsSpec(
            name="test_spec",
            version="1.0.0",
            columns=[Column(name="col1", type=yads_type)],
        )
        converter = DuckdbSqlConverter(
            ast_config=SqlglotConverterConfig(fallback_type=exp.DataType.Type.TEXT)
        )

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            ddl = converter.convert(spec, mode="coerce")

        # Assert warnings for unsupported types
        if expected_warning is not None:
            assert len(w) == 1
            assert issubclass(w[0].category, ValidationWarning)
            assert expected_warning in str(w[0].message)
        else:
            assert len(w) == 0

        # Assert converted SQL
        assert ddl.strip() == f"CREATE TABLE test_spec (col1 {expected_sql})"


# %% Dialect behavior
class TestDuckdbSqlConverterDialect:
    def test_convert_full_spec_matches_duckdb_fixture(self):
        spec = from_yaml_path("tests/fixtures/spec/valid/full_spec.yaml")
        converter = DuckdbSqlConverter()

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            ddl = converter.convert(spec, pretty=True)

        assert len(w) == 8
        assert all(issubclass(wi.category, ValidationWarning) for wi in w)
        messages = "\n".join(str(wi.message) for wi in w)
        assert (
            "Data type 'TIMESTAMPLTZ' is not supported for column 'c_timestamp_ltz'."
            in messages
        )
        assert "Data type 'VOID' is not supported for column 'c_void'." in messages
        assert (
            "Data type 'GEOGRAPHY' is not supported for column 'c_geography'." in messages
        )
        assert (
            "Parameterized 'GEOMETRY' is not supported for column 'c_geometry'."
            in messages
        )
        assert "Data type 'VARIANT' is not supported for column 'c_variant'." in messages
        assert (
            "GENERATED ALWAYS AS IDENTITY is not supported for column 'c_int32_identity'."
            in messages
        )
        assert "NULLS FIRST is not supported in PRIMARY KEY constraints." in messages
        assert "The data type will be replaced with 'TIMESTAMPTZ'." in messages
        assert "The data type will be replaced with 'TEXT'." in messages
        assert "The parameters will be removed." in messages
        assert "The NULLS FIRST attribute will be removed." in messages

        with open("tests/fixtures/sql/duckdb/full_spec.sql", "r") as f:
            expected_sql = f.read().strip()

        assert ddl.strip() == expected_sql


# %% Validation rules wiring
class TestDuckdbSqlConverterValidation:
    @pytest.mark.parametrize(
        "yads_type, original_type_sql, expected_sql",
        [
            ("timestampltz", "TIMESTAMPLTZ", "TIMESTAMPTZ"),
            ("void", "VOID", "TEXT"),
            ("geography", "GEOGRAPHY", "TEXT"),
            ("variant", "VARIANT", "TEXT"),
        ],
    )
    def test_coerce_mode_replaces_to_duckdb_supported_and_warns(
        self, yads_type: str, original_type_sql: str, expected_sql: str
    ):
        yaml_string = f"""
        name: my_db.my_table
        version: 1
        columns:
          - name: col1
            type: {yads_type}
        """
        spec = from_yaml_string(yaml_string)

        converter = DuckdbSqlConverter()
        with pytest.warns(
            UserWarning,
            match=f"Data type '{original_type_sql}' is not supported for column 'col1'.",
        ):
            ddl = converter.convert(spec, mode="coerce", pretty=True)

        expected_ddl = f"""CREATE TABLE my_db.my_table (
  col1 {expected_sql}
)"""
        assert ddl.strip() == expected_ddl

    def test_coerce_mode_removes_geometry_parameters_and_warns(self):
        yaml_string = """
        name: my_db.my_table
        version: 1
        columns:
          - name: col1
            type: geometry
            params:
              srid: 4326
        """
        spec = from_yaml_string(yaml_string)

        converter = DuckdbSqlConverter()
        with pytest.warns(
            UserWarning,
            match="Parameterized 'GEOMETRY' is not supported for column 'col1'.",
        ):
            ddl = converter.convert(spec, mode="coerce", pretty=True)

        expected_ddl = """CREATE TABLE my_db.my_table (
  col1 GEOMETRY
)"""
        assert ddl.strip() == expected_ddl

    @pytest.mark.parametrize(
        "yads_type, original_type_sql",
        [
            ("timestampltz", "TIMESTAMPLTZ"),
            ("void", "VOID"),
            ("geography", "GEOGRAPHY"),
            ("variant", "VARIANT"),
        ],
    )
    def test_raise_mode_raises_ast_validation_error(
        self, yads_type: str, original_type_sql: str
    ):
        yaml_string = f"""
        name: my_db.my_table
        version: 1
        columns:
          - name: col1
            type: {yads_type}
        """
        spec = from_yaml_string(yaml_string)

        converter = DuckdbSqlConverter()
        with pytest.raises(
            AstValidationError,
            match=f"Data type '{original_type_sql}' is not supported for column 'col1'.",
        ):
            converter.convert(spec, mode="raise")

    def test_raise_mode_raises_for_parameterized_geometry(self):
        yaml_string = """
        name: my_db.my_table
        version: 1
        columns:
          - name: col1
            type: geometry
            params:
              srid: 4326
        """
        spec = from_yaml_string(yaml_string)

        converter = DuckdbSqlConverter()
        with pytest.raises(
            AstValidationError,
            match="Parameterized 'GEOMETRY' is not supported for column 'col1'.",
        ):
            converter.convert(spec, mode="raise")



"""Constraint definitions for table-level and column-level data validation."""

from __future__ import annotations

import textwrap
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import cached_property
from typing import Any

from .exceptions import InvalidConstraintError


@dataclass(frozen=True)
class ForeignKeyReference:
    """Reference specification for foreign key constraints.

    Defines the target table and optional columns referenced by a foreign key.

    Args:
        table: Name of the referenced table.
        columns: List of column names in the referenced table.

    Raises:
        InvalidConstraintError: If columns is an empty list.
    """

    table: str
    columns: list[str] | None = None

    def __post_init__(self):
        if self.columns == []:
            raise InvalidConstraintError(
                "ForeignKeyReference 'columns' cannot be an empty list."
            )

    def __str__(self) -> str:
        if self.columns:
            return f"{self.table}({', '.join(self.columns)})"
        return self.table


# Column Constraints
class ColumnConstraint(ABC):
    """Abstract base class for column-level constraints.

    Applied to individual columns to define validation rules.
    """


@dataclass(frozen=True)
class NotNullConstraint(ColumnConstraint):
    """Constraint requiring that column values cannot be NULL."""

    def __str__(self) -> str:
        return "NotNullConstraint()"


@dataclass(frozen=True)
class PrimaryKeyConstraint(ColumnConstraint):
    """Constraint designating a column as the primary key."""

    def __str__(self) -> str:
        return "PrimaryKeyConstraint()"


@dataclass(frozen=True)
class DefaultConstraint(ColumnConstraint):
    """Constraint providing a default value for a column.

    Specifies the value to use when no explicit value is provided during insert
    operations. The default value should be compatible with the column's data type.

    Args:
        value: The default value to use.
    """

    value: Any

    def __str__(self) -> str:
        return f"DefaultConstraint(value={self.value!r})"


@dataclass(frozen=True)
class ForeignKeyConstraint(ColumnConstraint):
    """Column-level foreign key constraint.

    Args:
        references: The `ForeignKeyReference` specifying the target table and columns.
        name: Optional name for the constraint.

    Example:
        ```python
        # Simple foreign key to users table
        ForeignKeyConstraint(
            references=ForeignKeyReference(table="users")
        )

        # Named foreign key with specific column
        ForeignKeyConstraint(
            name="fk_order_customer",
            references=ForeignKeyReference(table="customers", columns=["id"])
        )
        ```
    """

    references: ForeignKeyReference
    name: str | None = None

    def __str__(self) -> str:
        parts: list[str] = []
        if self.name:
            parts.append(f"name={self.name!r}")
        parts.append(f"references={self.references}")

        pretty_parts = ", ".join(parts)
        return f"ForeignKeyConstraint({pretty_parts})"


@dataclass(frozen=True)
class IdentityConstraint(ColumnConstraint):
    """Constraint for auto-incrementing identity columns.

    Args:
        always: If True, values are always generated (GENERATED ALWAYS).
               If False, allows manual value insertion (GENERATED BY DEFAULT).
        start: Starting value for the sequence. If None, uses database default.
        increment: Increment step for each new value. If None, uses database default.

    Raises:
        InvalidConstraintError: If increment is zero.
    """

    always: bool = True
    start: int | None = None
    increment: int | None = None

    def __post_init__(self):
        if self.increment == 0:
            raise InvalidConstraintError(
                f"Identity 'increment' must be a non-zero integer, not {self.increment}."
            )


# Table Constraints
@dataclass(frozen=True)
class TableConstraint(ABC):
    """Abstract base class for table-level constraints."""

    @cached_property
    @abstractmethod
    def constrained_columns(self) -> list[str]:
        """Return the list of column names involved in this constraint."""


@dataclass(frozen=True)
class PrimaryKeyTableConstraint(TableConstraint):
    """Table-level primary key constraint.

    Args:
        columns: List of column names that form the composite primary key.
        name: Optional name for the constraint.

    Raises:
        InvalidConstraintError: If columns list is empty.

    Example:
        ```python
        # Composite primary key
        PrimaryKeyTableConstraint(
            columns=["order_id", "line_number"],
            name="pk_order_lines"
        )

        # Simple composite key without name
        PrimaryKeyTableConstraint(columns=["year", "month", "category"])
        ```
    """

    columns: list[str]
    name: str | None = None

    def __post_init__(self):
        if not self.columns:
            raise InvalidConstraintError(
                "PrimaryKeyTableConstraint 'columns' cannot be empty."
            )

    @cached_property
    def constrained_columns(self) -> list[str]:
        return self.columns

    def __str__(self) -> str:
        parts: list[str] = []
        if self.name:
            parts.append(f"name={self.name!r}")

        formatted_columns = ",\n".join(f"{col!r}" for col in self.columns)
        indented_columns = textwrap.indent(formatted_columns, "  ")
        parts.append(f"columns=[\n{indented_columns}\n]")

        pretty_parts = ",\n".join(parts)
        indented_parts = textwrap.indent(pretty_parts, "  ")
        return f"PrimaryKeyTableConstraint(\n{indented_parts}\n)"


@dataclass(frozen=True)
class ForeignKeyTableConstraint(TableConstraint):
    """Table-level foreign key constraint for composite relationships.

    Defines a foreign key across multiple columns. The number of columns must
    match the number of referenced columns.

    Args:
        columns: List of column names in this table that form the foreign key.
        references: The ForeignKeyReference specifying the target table and columns.
        name: Optional name for the constraint.

    Raises:
        InvalidConstraintError: If columns list is empty or if the number of
                              columns doesn't match the referenced columns.

    Example:
        ```python
        # Composite foreign key
        ForeignKeyTableConstraint(
            columns=["customer_id", "customer_region"],
            references=ForeignKeyReference(
                table="customers",
                columns=["id", "region"]
            ),
            name="fk_order_customer"
        )
        ```
    """

    columns: list[str]
    references: ForeignKeyReference
    name: str | None = None

    def __post_init__(self):
        if not self.columns:
            raise InvalidConstraintError(
                "ForeignKeyTableConstraint 'columns' cannot be empty."
            )
        if self.references.columns and len(self.columns) != len(self.references.columns):
            raise InvalidConstraintError(
                f"The number of columns in the foreign key ({len(self.columns)}) must match the number of "
                f"referenced columns ({len(self.references.columns)})."
            )

    @cached_property
    def constrained_columns(self) -> list[str]:
        return self.columns

    def __str__(self) -> str:
        parts: list[str] = []
        if self.name:
            parts.append(f"name={self.name!r}")

        formatted_columns = ",\n".join(f"{col!r}" for col in self.columns)
        indented_columns = textwrap.indent(formatted_columns, "  ")
        parts.append(f"columns=[\n{indented_columns}\n]")
        parts.append(f"references={self.references}")

        pretty_parts = ",\n".join(parts)
        indented_parts = textwrap.indent(pretty_parts, "  ")
        return f"ForeignKeyTableConstraint(\n{indented_parts}\n)"


CONSTRAINT_EQUIVALENTS: dict[type[ColumnConstraint], type[TableConstraint]] = {
    PrimaryKeyConstraint: PrimaryKeyTableConstraint,
    ForeignKeyConstraint: ForeignKeyTableConstraint,
}



"""Executable documentation examples."""

from .base import (
    ExampleBlockRequest,
    ExampleDefinition,
    ExampleCallable,
    ExampleBlockSource,
)

__all__ = [
    "ExampleBlockRequest",
    "ExampleDefinition",
    "ExampleCallable",
    "ExampleBlockSource",
]



#!/usr/bin/env python3
"""SQL Server integration test for SqlServerLoader.

This script validates that the SqlServerLoader can correctly introspect
SQL Server table schemas and convert them to YadsSpec instances.
"""

import sys

import pyodbc

from yads import types as ytypes
from yads.constraints import (
    DefaultConstraint,
    ForeignKeyConstraint,
    IdentityConstraint,
    NotNullConstraint,
    PrimaryKeyConstraint,
)
from yads.loaders.sql import SqlServerLoader


def get_connection() -> pyodbc.Connection:
    """Get a connection to the SQL Server test database."""
    connection_string = (
        "DRIVER={ODBC Driver 18 for SQL Server};"
        "SERVER=localhost,1433;"
        "DATABASE=yads_test;"
        "UID=sa;"
        "PWD=YadsTest123!;"
        "TrustServerCertificate=yes;"
    )
    return pyodbc.connect(connection_string)


def setup_test_tables(conn: pyodbc.Connection) -> None:
    """Create test tables with various SQL Server types and constraints."""
    cursor = conn.cursor()

    # Drop existing test tables
    cursor.execute("""
        IF OBJECT_ID('dbo.test_constraints', 'U') IS NOT NULL
            DROP TABLE dbo.test_constraints
    """)
    cursor.execute("""
        IF OBJECT_ID('dbo.test_types', 'U') IS NOT NULL
            DROP TABLE dbo.test_types
    """)
    cursor.execute("""
        IF OBJECT_ID('dbo.test_referenced', 'U') IS NOT NULL
            DROP TABLE dbo.test_referenced
    """)
    conn.commit()

    # Create referenced table for foreign key tests
    cursor.execute("""
        CREATE TABLE dbo.test_referenced (
            id BIGINT PRIMARY KEY,
            name NVARCHAR(100) NOT NULL
        )
    """)

    # Create table with various types
    cursor.execute("""
        CREATE TABLE dbo.test_types (
            -- Integer types
            col_tinyint TINYINT,
            col_smallint SMALLINT,
            col_int INT,
            col_bigint BIGINT,

            -- Floating point types
            col_real REAL,
            col_float FLOAT,
            col_decimal DECIMAL(10, 2),

            -- String types - ANSI
            col_char CHAR(10),
            col_varchar VARCHAR(255),
            col_varchar_max VARCHAR(MAX),
            col_text TEXT,

            -- String types - Unicode
            col_nchar NCHAR(10),
            col_nvarchar NVARCHAR(255),
            col_nvarchar_max NVARCHAR(MAX),
            col_ntext NTEXT,

            -- Binary types
            col_binary BINARY(16),
            col_varbinary VARBINARY(256),
            col_varbinary_max VARBINARY(MAX),
            col_image IMAGE,

            -- Boolean type
            col_bit BIT,

            -- Date/Time types
            col_date DATE,
            col_time TIME,
            col_smalldatetime SMALLDATETIME,
            col_datetime DATETIME,
            col_datetime2 DATETIME2,
            col_datetimeoffset DATETIMEOFFSET,

            -- UUID type
            col_uniqueidentifier UNIQUEIDENTIFIER,

            -- Spatial types
            col_geometry GEOMETRY,
            col_geography GEOGRAPHY
        )
    """)

    # Create table with constraints
    cursor.execute("""
        CREATE TABLE dbo.test_constraints (
            -- Primary key (single column)
            id BIGINT PRIMARY KEY,

            -- Identity column
            seq_id INT IDENTITY(1,1),

            -- Not null column
            required_col NVARCHAR(100) NOT NULL,

            -- Default values
            default_string NVARCHAR(50) DEFAULT 'hello',
            default_int INT DEFAULT 42,
            default_float FLOAT DEFAULT 3.14,
            default_null NVARCHAR(50) DEFAULT NULL,

            -- Foreign key
            ref_id BIGINT REFERENCES dbo.test_referenced(id),

            -- Computed column
            upper_required AS UPPER(required_col) PERSISTED
        )
    """)

    conn.commit()
    cursor.close()


def test_basic_type_conversion(conn: pyodbc.Connection) -> None:
    """Test basic type conversion from SQL Server to YadsTypes."""
    print("\n" + "-" * 60)
    print("Test: Basic Type Conversion")
    print("-" * 60)

    loader = SqlServerLoader(conn)
    spec = loader.load("test_types")

    print(f"  Loaded spec: {spec.name}")
    print(f"  Columns: {len(spec.columns)}")

    # Define expected types for each column
    expected_types: dict[str, ytypes.YadsType] = {
        # Integers
        "col_tinyint": ytypes.Integer(bits=8, signed=False),
        "col_smallint": ytypes.Integer(bits=16, signed=True),
        "col_int": ytypes.Integer(bits=32, signed=True),
        "col_bigint": ytypes.Integer(bits=64, signed=True),
        # Floats
        "col_real": ytypes.Float(bits=32),
        "col_float": ytypes.Float(bits=64),
        "col_decimal": ytypes.Decimal(precision=10, scale=2),
        # Strings - ANSI
        "col_char": ytypes.String(length=10),
        "col_varchar": ytypes.String(length=255),
        "col_varchar_max": ytypes.String(),
        "col_text": ytypes.String(),
        # Strings - Unicode
        "col_nchar": ytypes.String(length=10),
        "col_nvarchar": ytypes.String(length=255),
        "col_nvarchar_max": ytypes.String(),
        "col_ntext": ytypes.String(),
        # Binary
        "col_binary": ytypes.Binary(length=16),
        "col_varbinary": ytypes.Binary(length=256),
        "col_varbinary_max": ytypes.Binary(),
        "col_image": ytypes.Binary(),
        # Boolean
        "col_bit": ytypes.Boolean(),
        # Date/Time
        "col_date": ytypes.Date(bits=32),
        "col_time": ytypes.Time(unit=ytypes.TimeUnit.US),
        "col_smalldatetime": ytypes.TimestampNTZ(unit=ytypes.TimeUnit.S),
        "col_datetime": ytypes.TimestampNTZ(unit=ytypes.TimeUnit.MS),
        "col_datetime2": ytypes.TimestampNTZ(unit=ytypes.TimeUnit.US),
        "col_datetimeoffset": ytypes.TimestampTZ(unit=ytypes.TimeUnit.US, tz="UTC"),
        # UUID
        "col_uniqueidentifier": ytypes.UUID(),
        # Spatial
        "col_geometry": ytypes.Geometry(),
        "col_geography": ytypes.Geography(),
    }

    # Check expected types
    columns_by_name = {col.name: col for col in spec.columns}
    errors = []

    for col_name, expected_type in expected_types.items():
        if col_name not in columns_by_name:
            errors.append(f"  X Column '{col_name}' not found in spec")
            continue

        actual_type = columns_by_name[col_name].type
        if actual_type != expected_type:
            errors.append(
                f"  X Column '{col_name}': expected {expected_type}, got {actual_type}"
            )
        else:
            print(f"  + {col_name}: {actual_type}")

    if errors:
        for error in errors:
            print(error)
        raise AssertionError(f"Type conversion errors: {len(errors)}")

    print("+ All basic types converted correctly")


def test_constraints(conn: pyodbc.Connection) -> None:
    """Test constraint loading."""
    print("\n" + "-" * 60)
    print("Test: Constraint Loading")
    print("-" * 60)

    loader = SqlServerLoader(conn)
    spec = loader.load("test_constraints")

    columns_by_name = {col.name: col for col in spec.columns}

    # Test Primary Key
    col_id = columns_by_name["id"]
    pk_constraints = [
        c for c in col_id.constraints if isinstance(c, PrimaryKeyConstraint)
    ]
    assert len(pk_constraints) == 1, "Expected PRIMARY KEY constraint on 'id'"
    print("  + Primary key constraint detected")

    # Test Identity column
    col_seq_id = columns_by_name["seq_id"]
    identity_constraints = [
        c for c in col_seq_id.constraints if isinstance(c, IdentityConstraint)
    ]
    assert len(identity_constraints) == 1, "Expected IDENTITY constraint on 'seq_id'"
    assert identity_constraints[0].start == 1
    assert identity_constraints[0].increment == 1
    print("  + Identity constraint detected (seed=1, increment=1)")

    # Test NOT NULL
    col_required = columns_by_name["required_col"]
    not_null_constraints = [
        c for c in col_required.constraints if isinstance(c, NotNullConstraint)
    ]
    assert len(not_null_constraints) == 1, (
        "Expected NOT NULL constraint on 'required_col'"
    )
    print("  + NOT NULL constraint detected")

    # Test DEFAULT values
    col_default_string = columns_by_name["default_string"]
    default_constraints = [
        c for c in col_default_string.constraints if isinstance(c, DefaultConstraint)
    ]
    assert len(default_constraints) == 1, (
        "Expected DEFAULT constraint on 'default_string'"
    )
    assert default_constraints[0].value == "hello"
    print("  + DEFAULT string value detected")

    col_default_int = columns_by_name["default_int"]
    default_int_constraints = [
        c for c in col_default_int.constraints if isinstance(c, DefaultConstraint)
    ]
    assert len(default_int_constraints) == 1
    assert default_int_constraints[0].value == 42
    print("  + DEFAULT integer value detected")

    col_default_float = columns_by_name["default_float"]
    default_float_constraints = [
        c for c in col_default_float.constraints if isinstance(c, DefaultConstraint)
    ]
    assert len(default_float_constraints) == 1
    assert default_float_constraints[0].value == 3.14
    print("  + DEFAULT float value detected")

    # Test Foreign Key
    col_ref_id = columns_by_name["ref_id"]
    fk_constraints = [
        c for c in col_ref_id.constraints if isinstance(c, ForeignKeyConstraint)
    ]
    assert len(fk_constraints) == 1, "Expected FOREIGN KEY constraint on 'ref_id'"
    assert fk_constraints[0].references.table == "test_referenced"
    print("  + Foreign key constraint detected")

    print("+ All constraints loaded correctly")


def test_computed_columns(conn: pyodbc.Connection) -> None:
    """Test computed column handling."""
    print("\n" + "-" * 60)
    print("Test: Computed Columns")
    print("-" * 60)

    loader = SqlServerLoader(conn)
    spec = loader.load("test_constraints")

    columns_by_name = {col.name: col for col in spec.columns}

    col_upper = columns_by_name["upper_required"]
    assert col_upper.generated_as is not None, (
        "Expected generated_as for 'upper_required'"
    )
    assert col_upper.generated_as.column == "required_col"
    assert col_upper.generated_as.transform == "upper"
    print(f"  + Computed column detected: {col_upper.generated_as}")

    print("+ Computed columns handled correctly")


def test_spec_metadata(conn: pyodbc.Connection) -> None:
    """Test spec metadata generation."""
    print("\n" + "-" * 60)
    print("Test: Spec Metadata")
    print("-" * 60)

    loader = SqlServerLoader(conn)

    # Test default name (fully qualified: catalog.schema.table)
    spec = loader.load("test_types")
    assert spec.name == "yads_test.dbo.test_types"
    print(f"  + Default name: {spec.name}")

    # Test custom name
    spec = loader.load("test_types", name="custom.schema.test")
    assert spec.name == "custom.schema.test"
    print(f"  + Custom name: {spec.name}")

    # Test version
    spec = loader.load("test_types", version=5)
    assert spec.version == 5
    print(f"  + Version: {spec.version}")

    # Test description
    spec = loader.load("test_types", description="Test table for types")
    assert spec.description == "Test table for types"
    print(f"  + Description: {spec.description}")

    print("+ Spec metadata generated correctly")


def main() -> int:
    """Run all SQL Server integration tests."""
    print("=" * 60)
    print("SQL Server Loader Integration Test")
    print("=" * 60)

    print("\n> Connecting to SQL Server...")
    try:
        conn = get_connection()
        print("+ Connected to SQL Server")
    except Exception as e:
        print(f"X Failed to connect: {e}")
        return 1

    try:
        print("\n> Setting up test tables...")
        setup_test_tables(conn)
        print("+ Test tables created")

        # Run all tests
        test_basic_type_conversion(conn)
        test_constraints(conn)
        test_computed_columns(conn)
        test_spec_metadata(conn)

        print("\n" + "=" * 60)
        print("+ All SQL Server integration tests PASSED")
        print("=" * 60)
        return 0

    except Exception as e:
        print(f"\nX Test FAILED: {e}")
        import traceback

        traceback.print_exc()
        return 1

    finally:
        conn.close()


if __name__ == "__main__":
    sys.exit(main())



"""Executable examples for loading specs from YAML strings."""

from __future__ import annotations

from ..base import ExampleBlockRequest, ExampleDefinition


def _yaml_loader_lowlevel_example() -> None:
    from yads.loaders import YamlLoader

    yaml_string = """
name: "prod.assessments.submissions"
version: 1
yads_spec_version: "0.0.2"

columns:
  - name: "submission_id"
    type: "bigint"
    constraints:
      primary_key: true
      not_null: true

  - name: "completion_percent"
    type: "decimal"
    params:
      precision: 5
      scale: 2
    constraints:
      default: 0.00

  - name: "time_taken_seconds"
    type: "integer"

  - name: "submitted_at"
    type: "timestamptz"
    params:
      tz: "UTC"
    """

    loader = YamlLoader()
    spec = loader.load(yaml_string)
    print(spec)


EXAMPLE = ExampleDefinition(
    example_id="yaml-loader-basic",
    blocks=(
        ExampleBlockRequest(
            slug="loader-example-lowlevel-code",
            language="python",
            source="callable",
            callable=_yaml_loader_lowlevel_example,
        ),
        ExampleBlockRequest(
            slug="loader-example-lowlevel-output",
            language="text",
            source="stdout",
            callable=_yaml_loader_lowlevel_example,
        ),
    ),
)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .sql_converter import (
        SqlConverter,
        SqlConverterConfig,
        SparkSqlConverter,
        DuckdbSqlConverter,
    )
    from .ast_converter import AstConverter, SqlglotConverter, SqlglotConverterConfig
    from .validators.ast_validator import AstValidator
    from .validators.ast_validation_rules import (
        AstValidationRule,
        DisallowType,
        DisallowUserDefinedType,
        DisallowFixedLengthString,
        DisallowFixedLengthBinary,
        DisallowNegativeScaleDecimal,
        DisallowParameterizedGeometry,
        DisallowColumnConstraintGeneratedIdentity,
        DisallowTableConstraintPrimaryKeyNullsFirst,
    )

__all__ = [
    "AstConverter",
    "SqlConverter",
    "SqlConverterConfig",
    "SparkSqlConverter",
    "DuckdbSqlConverter",
    "SqlglotConverter",
    "SqlglotConverterConfig",
    "AstValidator",
    "AstValidationRule",
    "DisallowType",
    "DisallowUserDefinedType",
    "DisallowFixedLengthString",
    "DisallowFixedLengthBinary",
    "DisallowNegativeScaleDecimal",
    "DisallowParameterizedGeometry",
    "DisallowColumnConstraintGeneratedIdentity",
    "DisallowTableConstraintPrimaryKeyNullsFirst",
]


def __getattr__(name: str) -> Any:
    """Lazy import SQL converters to avoid eager sqlglot dependency."""
    if name in (
        "SqlConverter",
        "SqlConverterConfig",
        "SparkSqlConverter",
        "DuckdbSqlConverter",
    ):
        from . import sql_converter

        return getattr(sql_converter, name)
    if name in ("AstConverter", "SqlglotConverter", "SqlglotConverterConfig"):
        from . import ast_converter

        return getattr(ast_converter, name)
    if name == "AstValidator":
        from .validators import ast_validator

        return getattr(ast_validator, name)
    if name in (
        "AstValidationRule",
        "DisallowType",
        "DisallowUserDefinedType",
        "DisallowFixedLengthString",
        "DisallowFixedLengthBinary",
        "DisallowNegativeScaleDecimal",
        "DisallowParameterizedGeometry",
        "DisallowColumnConstraintGeneratedIdentity",
        "DisallowTableConstraintPrimaryKeyNullsFirst",
    ):
        from .validators import ast_validation_rules

        return getattr(ast_validation_rules, name)
    raise AttributeError(name)



"""Unit tests for PolarsLoader."""

import warnings
import polars as pl  # type: ignore[import-untyped]
import pytest

from yads.exceptions import LoaderConfigError, UnsupportedFeatureError, ValidationWarning
from yads.loaders import PolarsLoader, PolarsLoaderConfig
from yads.types import (
    YadsType,
    String,
    Integer,
    Float,
    Decimal,
    Boolean,
    Binary,
    Date,
    TimeUnit,
    Time,
    Timestamp,
    TimestampTZ,
    Duration,
    Array,
    Struct,
    Void,
    Variant,
    Tensor,
)


# fmt: off
# %% Type conversion tests
class TestPolarsLoaderTypeConversion:
    @pytest.mark.parametrize(
        "pl_type, expected_yads_type",
        [
            # Null / Boolean
            (pl.Null, Void()),
            (pl.Boolean, Boolean()),
            
            # Integers
            (pl.Int8, Integer(bits=8, signed=True)),
            (pl.Int16, Integer(bits=16, signed=True)),
            (pl.Int32, Integer(bits=32, signed=True)),
            (pl.Int64, Integer(bits=64, signed=True)),
            (pl.UInt8, Integer(bits=8, signed=False)),
            (pl.UInt16, Integer(bits=16, signed=False)),
            (pl.UInt32, Integer(bits=32, signed=False)),
            (pl.UInt64, Integer(bits=64, signed=False)),
            
            # Floats
            (pl.Float32, Float(bits=32)),
            (pl.Float64, Float(bits=64)),
            
            # Strings / Binary
            (pl.String, String()),
            (pl.Utf8, String()),  # Alias for String
            (pl.Binary, Binary()),
            
            # Date / Time
            (pl.Date, Date(bits=32)),
            (pl.Time, Time(unit=TimeUnit.NS, bits=64)),
            
            # Decimal
            (pl.Decimal(precision=10, scale=2), Decimal(precision=10, scale=2)),
            (pl.Decimal(precision=20, scale=3), Decimal(precision=20, scale=3)),
        ],
    )
    def test_convert_primitive_types(
        self, pl_type: pl.DataType, expected_yads_type: YadsType
    ):
        schema = pl.Schema({"col1": pl_type})
        loader = PolarsLoader()
        spec = loader.load(schema, name="test_spec", version=1)
        
        assert spec.name == "test_spec"
        assert spec.version == 1
        assert len(spec.columns) == 1
        
        column = spec.columns[0]
        assert column.name == "col1"
        assert column.type == expected_yads_type
        assert column.is_nullable is True  # Polars Schema doesn't track nullability

    def test_decimal_without_explicit_scale_defaults(self):
        """Decimals without an explicit scale should fall back to generic Decimal."""
        schema = pl.Schema({"col1": pl.Decimal(precision=38, scale=None)})
        loader = PolarsLoader()
        spec = loader.load(schema, name="test_spec", version=1)

        column = spec.columns[0]
        assert column.name == "col1"
        assert column.type == Decimal()

    @pytest.mark.parametrize(
        "pl_type, expected_yads_type",
        [
            # Timestamp without timezone
            (pl.Datetime(time_unit="ms", time_zone=None), Timestamp(unit=TimeUnit.MS)),
            (pl.Datetime(time_unit="us", time_zone=None), Timestamp(unit=TimeUnit.US)),
            (pl.Datetime(time_unit="ns", time_zone=None), Timestamp(unit=TimeUnit.NS)),
            
            # Timestamp with timezone
            (pl.Datetime(time_unit="ms", time_zone="UTC"), TimestampTZ(unit=TimeUnit.MS, tz="UTC")),
            (pl.Datetime(time_unit="us", time_zone="America/New_York"), TimestampTZ(unit=TimeUnit.US, tz="America/New_York")),
            (pl.Datetime(time_unit="ns", time_zone="Europe/London"), TimestampTZ(unit=TimeUnit.NS, tz="Europe/London")),
            
            # Duration
            (pl.Duration(time_unit="ms"), Duration(unit=TimeUnit.MS)),
            (pl.Duration(time_unit="us"), Duration(unit=TimeUnit.US)),
            (pl.Duration(time_unit="ns"), Duration(unit=TimeUnit.NS)),
        ],
    )
    def test_convert_temporal_types(
        self, pl_type: pl.DataType, expected_yads_type: YadsType
    ):
        schema = pl.Schema({"col1": pl_type})
        loader = PolarsLoader()
        spec = loader.load(schema, name="test_spec", version=1)
        
        column = spec.columns[0]
        assert column.name == "col1"
        assert column.type == expected_yads_type
        assert column.is_nullable is True

    @pytest.mark.parametrize(
        "pl_type, expected_element_type, expected_size",
        [
            # Variable-length lists
            (pl.List(pl.String), String(), None),
            (pl.List(pl.Int32), Integer(bits=32, signed=True), None),
            (pl.List(pl.Float64), Float(bits=64), None),
            (pl.List(pl.Boolean), Boolean(), None),
            
            # Fixed-size arrays (1D)
            (pl.Array(pl.String, shape=5), String(), 5),
            (pl.Array(pl.Int32, shape=10), Integer(bits=32, signed=True), 10),
            (pl.Array(pl.Float64, shape=3), Float(bits=64), 3),
        ],
    )
    def test_convert_list_and_array_types(
        self, pl_type: pl.DataType, expected_element_type: YadsType, expected_size: int | None
    ):
        schema = pl.Schema({"col1": pl_type})
        loader = PolarsLoader()
        spec = loader.load(schema, name="test_spec", version=1)
        
        column = spec.columns[0]
        assert isinstance(column.type, Array)
        assert column.type.element == expected_element_type
        assert column.type.size == expected_size

    @pytest.mark.parametrize(
        "pl_type, expected_element_type, expected_shape",
        [
            # Multi-dimensional arrays -> Tensor
            (pl.Array(pl.Int32, shape=(10, 20)), Integer(bits=32, signed=True), (10, 20)),
            (pl.Array(pl.Float64, shape=(5, 10, 15)), Float(bits=64), (5, 10, 15)),
            (pl.Array(pl.String, shape=(100, 50)), String(), (100, 50)),
            (pl.Array(pl.Boolean, shape=(3, 4, 5, 6)), Boolean(), (3, 4, 5, 6)),
        ],
    )
    def test_convert_multidimensional_array_to_tensor(
        self, pl_type: pl.DataType, expected_element_type: YadsType, expected_shape: tuple[int, ...]
    ):
        schema = pl.Schema({"col1": pl_type})
        loader = PolarsLoader()
        spec = loader.load(schema, name="test_spec", version=1)
        
        column = spec.columns[0]
        assert isinstance(column.type, Tensor)
        assert column.type.element == expected_element_type
        assert column.type.shape == expected_shape

    def test_convert_struct_type(self):
        schema = pl.Schema({
            "struct_col": pl.Struct([
                pl.Field("x", pl.Int32),
                pl.Field("y", pl.String),
                pl.Field("z", pl.Float64),
            ])
        })
        loader = PolarsLoader()
        spec = loader.load(schema, name="test_spec", version=1)
        
        column = spec.columns[0]
        assert isinstance(column.type, Struct)
        assert len(column.type.fields) == 3
        
        field_x = column.type.fields[0]
        assert field_x.name == "x"
        assert field_x.type == Integer(bits=32, signed=True)
        assert field_x.is_nullable is True
        
        field_y = column.type.fields[1]
        assert field_y.name == "y"
        assert field_y.type == String()
        assert field_y.is_nullable is True
        
        field_z = column.type.fields[2]
        assert field_z.name == "z"
        assert field_z.type == Float(bits=64)
        assert field_z.is_nullable is True

    def test_convert_nested_complex_types(self):
        inner_struct = pl.Struct([
            pl.Field("id", pl.Int32),
            pl.Field("data", pl.List(pl.String)),
        ])
        schema = pl.Schema({"nested": pl.List(inner_struct)})
        loader = PolarsLoader()
        spec = loader.load(schema, name="test_spec", version=1)
        
        column = spec.columns[0]
        assert isinstance(column.type, Array)
        
        element_type = column.type.element
        assert isinstance(element_type, Struct)
        assert len(element_type.fields) == 2
        
        id_field = element_type.fields[0]
        assert id_field.name == "id"
        assert id_field.type == Integer(bits=32, signed=True)
        
        data_field = element_type.fields[1]
        assert data_field.name == "data"
        assert isinstance(data_field.type, Array)
        assert data_field.type.element == String()

    def test_convert_deeply_nested_complex_types(self):
        inner_struct = pl.Struct([
            pl.Field("id", pl.Int32),
            pl.Field("values", pl.Array(pl.Float64, shape=5)),
        ])
        schema = pl.Schema({
            "complex_col": pl.List(inner_struct)
        })
        loader = PolarsLoader()
        spec = loader.load(schema, name="test_spec", version=1)
        
        column = spec.columns[0]
        assert isinstance(column.type, Array)
        
        # Element should be Struct
        element_type = column.type.element
        assert isinstance(element_type, Struct)
        
        struct_fields = element_type.fields
        assert len(struct_fields) == 2
        assert struct_fields[0].name == "id"
        assert struct_fields[0].type == Integer(bits=32, signed=True)
        assert struct_fields[1].name == "values"
        assert isinstance(struct_fields[1].type, Array)
        assert struct_fields[1].type.element == Float(bits=64)
        assert struct_fields[1].type.size == 5

    @pytest.mark.skipif(
        not hasattr(pl, "Object"),
        reason="pl.Object not available in this polars version"
    )
    def test_convert_object_type(self):
        """Test that pl.Object maps to Variant."""
        schema = pl.Schema({"obj_col": pl.Object})
        loader = PolarsLoader()
        spec = loader.load(schema, name="test_spec", version=1)
        
        column = spec.columns[0]
        assert column.name == "obj_col"
        assert isinstance(column.type, Variant)

    def test_convert_nested_array_with_struct(self):
        """Test nested structure with arrays containing structs with arrays."""
        inner_struct = pl.Struct([
            pl.Field("name", pl.String),
            pl.Field("tags", pl.List(pl.String)),
            pl.Field("matrix", pl.Array(pl.Int32, shape=(3, 3))),
        ])
        schema = pl.Schema({
            "records": pl.List(inner_struct)
        })
        loader = PolarsLoader()
        spec = loader.load(schema, name="test_spec", version=1)
        
        column = spec.columns[0]
        assert isinstance(column.type, Array)
        assert isinstance(column.type.element, Struct)
        
        fields = column.type.element.fields
        assert len(fields) == 3
        
        # Check name field
        assert fields[0].name == "name"
        assert fields[0].type == String()
        
        # Check tags field (variable-length list)
        assert fields[1].name == "tags"
        assert isinstance(fields[1].type, Array)
        assert fields[1].type.element == String()
        assert fields[1].type.size is None
        
        # Check matrix field (2D fixed array -> Tensor)
        assert fields[2].name == "matrix"
        assert isinstance(fields[2].type, Tensor)
        assert fields[2].type.element == Integer(bits=32, signed=True)
        assert fields[2].type.shape == (3, 3)
# fmt: on


# %% Field nullability tests
class TestPolarsLoaderNullability:
    def test_all_fields_nullable_by_default(self):
        """Polars Schema doesn't track nullability, so all fields are nullable."""
        schema = pl.Schema(
            {
                "col1": pl.String,
                "col2": pl.Int32,
            }
        )
        loader = PolarsLoader()
        spec = loader.load(schema, name="test_spec", version=1)

        for column in spec.columns:
            assert column.is_nullable is True
            assert len(column.constraints) == 0

    def test_nested_fields_nullable(self):
        """Nested struct fields should also be nullable by default."""
        schema = pl.Schema(
            {
                "struct_col": pl.Struct(
                    [
                        pl.Field("x", pl.Int32),
                        pl.Field("y", pl.String),
                    ]
                )
            }
        )
        loader = PolarsLoader()
        spec = loader.load(schema, name="test_spec", version=1)

        struct_col = spec.columns[0]
        assert isinstance(struct_col.type, Struct)

        for field in struct_col.type.fields:
            assert field.is_nullable is True
            assert len(field.constraints) == 0


# %% Schema-level tests
class TestPolarsLoaderSchema:
    def test_schema_without_description(self):
        schema = pl.Schema({"id": pl.Int32})
        loader = PolarsLoader()
        spec = loader.load(schema, name="test", version=1)

        assert spec.name == "test"
        assert spec.version == 1
        assert spec.description is None

    def test_schema_with_description(self):
        schema = pl.Schema({"id": pl.Int32})
        loader = PolarsLoader()
        spec = loader.load(schema, name="test", version=1, description="Test description")

        assert spec.name == "test"
        assert spec.version == 1
        assert spec.description == "Test description"

    def test_empty_schema(self):
        schema = pl.Schema({})
        loader = PolarsLoader()
        spec = loader.load(schema, name="empty", version=1)

        assert spec.name == "empty"
        assert spec.version == 1
        assert len(spec.columns) == 0

    def test_multiple_columns(self):
        schema = pl.Schema(
            {
                "id": pl.Int64,
                "name": pl.String,
                "age": pl.Int32,
                "score": pl.Float64,
            }
        )
        loader = PolarsLoader()
        spec = loader.load(schema, name="test", version=1)

        assert len(spec.columns) == 4
        assert spec.columns[0].name == "id"
        assert spec.columns[1].name == "name"
        assert spec.columns[2].name == "age"
        assert spec.columns[3].name == "score"


# %% Unsupported types and error handling
class TestPolarsLoaderUnsupportedTypes:
    def test_categorical_type_raises_error(self):
        """Test that Categorical type raises error in raise mode."""
        # Create a DataFrame with categorical data to get the schema
        df = pl.DataFrame({"cat_col": pl.Series(["a", "b", "c"], dtype=pl.Categorical)})
        schema = df.schema

        config = PolarsLoaderConfig(mode="raise")
        loader = PolarsLoader(config)

        with pytest.raises(
            UnsupportedFeatureError,
            match="PolarsLoader does not support Polars type.*for 'cat_col'",
        ):
            loader.load(schema, name="test", version=1)

    def test_enum_type_raises_error(self):
        """Test that Enum type raises error in raise mode."""
        if hasattr(pl, "Enum"):
            schema = pl.Schema({"enum_col": pl.Enum(["a", "b", "c"])})
            config = PolarsLoaderConfig(mode="raise")
            loader = PolarsLoader(config)

            with pytest.raises(
                UnsupportedFeatureError,
                match="PolarsLoader does not support Polars type.*for 'enum_col'",
            ):
                loader.load(schema, name="test", version=1)


# %% Configuration tests
class TestPolarsLoaderConfig:
    def test_default_config(self):
        config = PolarsLoaderConfig()
        assert config.mode == "coerce"
        assert config.fallback_type is None

    def test_custom_config(self):
        config = PolarsLoaderConfig(mode="raise", fallback_type=Binary())
        assert config.mode == "raise"
        assert config.fallback_type == Binary()

    def test_invalid_mode_raises_error(self):
        with pytest.raises(
            LoaderConfigError, match="mode must be one of 'raise' or 'coerce'"
        ):
            PolarsLoaderConfig(mode="invalid")  # type: ignore[arg-type]

    def test_invalid_fallback_type_raises_error(self):
        with pytest.raises(
            LoaderConfigError, match="fallback_type must be either String or Binary"
        ):
            PolarsLoaderConfig(fallback_type=Integer())  # type: ignore[arg-type]

    def test_config_is_immutable(self):
        config = PolarsLoaderConfig(mode="raise")
        # Config should be frozen
        with pytest.raises(AttributeError):
            config.mode = "coerce"  # type: ignore[misc]


class TestPolarsLoaderWithConfig:
    def test_loader_with_default_config(self):
        loader = PolarsLoader()
        assert loader.config.mode == "coerce"
        assert loader.config.fallback_type is None

    def test_loader_with_custom_config(self):
        config = PolarsLoaderConfig(mode="raise", fallback_type=Binary())
        loader = PolarsLoader(config)
        assert loader.config.mode == "raise"
        assert loader.config.fallback_type == Binary()

    def test_mode_override_in_load_method(self):
        config = PolarsLoaderConfig(mode="coerce", fallback_type=String())
        loader = PolarsLoader(config)

        # Create a DataFrame with categorical data to get the schema
        df = pl.DataFrame({"cat_col": pl.Series(["a", "b", "c"], dtype=pl.Categorical)})
        schema = df.schema

        with pytest.raises(UnsupportedFeatureError):
            loader.load(schema, name="test", version=1, mode="raise")

    def test_coercion_mode_without_fallback_raises(self):
        """Test that coerce mode raises when fallback_type is None."""
        config = PolarsLoaderConfig(mode="coerce")  # fallback_type=None
        loader = PolarsLoader(config)

        # Create a DataFrame with categorical data to get the schema
        df = pl.DataFrame({"cat_col": pl.Series(["a", "b", "c"], dtype=pl.Categorical)})
        schema = df.schema

        with pytest.raises(
            UnsupportedFeatureError,
            match="Specify a fallback_type to enable coercion",
        ):
            loader.load(schema, name="test", version=1)

    def test_coercion_mode_with_string_fallback(self):
        config = PolarsLoaderConfig(mode="coerce", fallback_type=String())
        loader = PolarsLoader(config)

        # Create a DataFrame with categorical data to get the schema
        df = pl.DataFrame({"cat_col": pl.Series(["a", "b", "c"], dtype=pl.Categorical)})
        schema = df.schema

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            spec = loader.load(schema, name="test", version=1)

            assert len(w) == 1
            assert "PolarsLoader does not support Polars type" in str(w[0].message)
            assert "for 'cat_col'" in str(w[0].message)
            assert "The data type will be coerced to string" in str(w[0].message)

        column = spec.columns[0]
        assert column.name == "cat_col"
        assert column.type == String()

    def test_coercion_mode_with_custom_binary_fallback(self):
        config = PolarsLoaderConfig(mode="coerce", fallback_type=Binary(length=10))
        loader = PolarsLoader(config)

        # Create a DataFrame with categorical data to get the schema
        df = pl.DataFrame({"cat_col": pl.Series(["a", "b", "c"], dtype=pl.Categorical)})
        schema = df.schema

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            spec = loader.load(schema, name="test", version=1)

            assert len(w) == 1
            assert "PolarsLoader does not support Polars type" in str(w[0].message)
            assert "for 'cat_col'" in str(w[0].message)
            assert "The data type will be coerced to binary(length=10)" in str(
                w[0].message
            )

        column = spec.columns[0]
        assert column.name == "cat_col"
        assert column.type == Binary(length=10)

    def test_raise_mode_with_unsupported_types(self):
        config = PolarsLoaderConfig(mode="raise")
        loader = PolarsLoader(config)

        # Create a DataFrame with categorical data to get the schema
        df = pl.DataFrame({"cat_col": pl.Series(["a", "b", "c"], dtype=pl.Categorical)})
        schema = df.schema

        with pytest.raises(
            UnsupportedFeatureError, match="PolarsLoader does not support Polars type"
        ):
            loader.load(schema, name="test", version=1)

    def test_multiple_unsupported_types_coercion(self):
        config = PolarsLoaderConfig(mode="coerce", fallback_type=String())
        loader = PolarsLoader(config)

        # Create a DataFrame with categorical data to get the schema
        df = pl.DataFrame(
            {
                "cat_col": pl.Series(["a", "b", "c"], dtype=pl.Categorical),
                "normal_col": pl.Series(["x", "y", "z"], dtype=pl.String),
            }
        )
        schema = df.schema

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            spec = loader.load(schema, name="test", version=1)

            assert len(w) == 1
            assert "PolarsLoader does not support Polars type" in str(w[0].message)
            assert "for 'cat_col'" in str(w[0].message)

        assert len(spec.columns) == 2
        assert spec.columns[0].name == "cat_col"
        assert spec.columns[0].type == String()  # coerced
        assert spec.columns[1].name == "normal_col"
        assert spec.columns[1].type == String()  # normal conversion

    def test_field_context_in_error_messages(self):
        config = PolarsLoaderConfig(mode="raise")
        loader = PolarsLoader(config)

        # Create a DataFrame with categorical data to get the schema
        df = pl.DataFrame({"my_field": pl.Series(["a", "b", "c"], dtype=pl.Categorical)})
        schema = df.schema

        with pytest.raises(UnsupportedFeatureError) as exc_info:
            loader.load(schema, name="test", version=1)

        assert "for 'my_field'" in str(exc_info.value)

    def test_nested_unsupported_types_coercion(self):
        config = PolarsLoaderConfig(mode="coerce", fallback_type=String())
        loader = PolarsLoader(config)

        df = pl.DataFrame(
            {
                "id": [1, 2],
                "cat_field": pl.Series(["a", "b"], dtype=pl.Categorical),
            }
        ).select(pl.struct(["id", "cat_field"]).alias("nested"))
        schema = df.schema

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            spec = loader.load(schema, name="test", version=1)

            assert len(w) == 1
            assert "PolarsLoader does not support Polars type" in str(w[0].message)
            assert "for 'cat_field'" in str(w[0].message)

        # Check that only the unsupported field within the struct was coerced
        column = spec.columns[0]
        assert column.name == "nested"
        assert isinstance(column.type, Struct)  # struct preserved
        assert len(column.type.fields) == 2

        # Check individual fields
        id_field = next(f for f in column.type.fields if f.name == "id")
        assert isinstance(id_field.type, Integer)  # normal field preserved

        cat_field = next(f for f in column.type.fields if f.name == "cat_field")
        assert isinstance(cat_field.type, String)  # unsupported field coerced

    def test_complex_nested_fallback_behavior(self):
        config = PolarsLoaderConfig(mode="coerce", fallback_type=String())
        loader = PolarsLoader(config)

        inner_struct = pl.Struct(
            [
                pl.Field("id", pl.Int32),
                pl.Field("unsupported_field", pl.Categorical),
                pl.Field("normal_string", pl.String),
            ]
        )

        array_of_structs = pl.List(inner_struct)

        schema = pl.Schema(
            {
                "id": pl.Int32,
                "complex_data": array_of_structs,
            }
        )

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            spec = loader.load(schema, name="test_complex", version=1)

        # Should have warnings for the unsupported type
        assert len(w) == 1
        assert issubclass(w[0].category, ValidationWarning)

        warning_message = str(w[0].message)
        assert "PolarsLoader does not support Polars type" in warning_message
        assert "unsupported_field" in warning_message

        # Check that the structure is preserved
        assert len(spec.columns) == 2

        # Normal field should be preserved
        id_col = spec.columns[0]
        assert id_col.name == "id"
        assert isinstance(id_col.type, Integer)

        # Complex field should still be an array
        complex_col = spec.columns[1]
        assert complex_col.name == "complex_data"
        assert isinstance(complex_col.type, Array)

        # Array element should still be a struct
        assert isinstance(complex_col.type.element, Struct)
        assert len(complex_col.type.element.fields) == 3

        # Check struct fields
        struct_fields = complex_col.type.element.fields
        id_field = next(f for f in struct_fields if f.name == "id")
        assert isinstance(id_field.type, Integer)  # preserved

        unsupported_field = next(
            f for f in struct_fields if f.name == "unsupported_field"
        )
        assert isinstance(unsupported_field.type, String)  # coerced

        normal_string_field = next(f for f in struct_fields if f.name == "normal_string")
        assert isinstance(normal_string_field.type, String)  # preserved


class TestPolarsLoaderHelpers:
    def test_normalize_time_unit_passthrough(self):
        assert PolarsLoader._normalize_time_unit(TimeUnit.MS) == TimeUnit.MS

    def test_normalize_time_unit_defaults_to_ns(self):
        assert PolarsLoader._normalize_time_unit(None) == TimeUnit.NS

    def test_extract_duration_unit_without_attribute_defaults_to_ns(self):
        class DummyDuration:
            time_unit = None

        assert PolarsLoader._extract_duration_unit(DummyDuration()) == "ns"

    def test_extract_datetime_helpers_handle_missing_values(self):
        class DummyDatetime:
            time_unit = None
            time_zone = None

        assert PolarsLoader._extract_datetime_unit(DummyDatetime()) == "ns"
        assert PolarsLoader._extract_datetime_timezone(DummyDatetime()) is None
